#  wfield - tools to analyse widefield data - NCAAS gui
# Copyright (C) 2020 Joao Couto - jpcouto@gmail.com
#
#  This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 

from .ncaas_utils import *
from PyQt5.QtWidgets import (QApplication,
                             QWidget,
                             QMainWindow,
                             QDockWidget,
                             QFormLayout,
                             QHBoxLayout,
                             QGridLayout,
                             QVBoxLayout,
                             QPushButton,
                             QGridLayout,
                             QTreeWidgetItem,
                             QTreeView,
                             QTextEdit,
                             QPlainTextEdit,
                             QLineEdit,
                             QScrollArea,
                             QCheckBox,
                             QComboBox,
                             QListWidget,
                             QLabel,
                             QProgressBar,
                             QFileDialog,
                             QMessageBox,
                             QDesktopWidget,
                             QListWidgetItem,
                             QFileSystemModel,
                             QAbstractItemView,
                             QTabWidget,
                             QMenu,
                             QDialog,
                             QDialogButtonBox,
                             QAction)
from functools import partial
try:
    from PyQt5.QtWebEngineWidgets import QWebEngineView
except:
    print('Could not load QWebEngineView.')
from PyQt5 import QtCore
from PyQt5.QtGui import QStandardItem, QStandardItemModel,QColor
from PyQt5.QtCore import Qt, QTimer,QMimeData

PMD_BUCKET = 'wfield-preprocess'
NMF_BUCKET = 'cshl-wfield-locanmf'

tempfile = pjoin(os.path.expanduser('~'),'.wfield','tempfile')

analysis_extensions = ['.bin','.tiff','.tif','.npy','.json','.dat']
ZIP_PMD = True

analysis_selection_dict = [dict(acronym = 'motion',
                                name='Motion correction',
                                desc = 'Performs the registration of frames to a reference taken from aligning the second 60 frames of the dataset to themselves.',
                                selected = True),
                           dict(acronym='compression',
                                name='Compression and denoising using PMD',
                                desc = 'Performs Penalized Matrix Decomposition (PMD) to compress and denoise the dataset. This step operates on motion corrected data.',
                                selected=True),
                           dict(acronym='hemodynamics_compensation',
                                name='Hemodynamics compensation',
                                desc = 'Performs hemodynamics compensation by regressing the second channel (isobestic for the indicator) into the first (indicator fluorescence) and subtracting the result to the first channel. This step requires a compressed dataset with 2 channels.',
                                selected = True),
                           dict(acronym='locaNMF',
                                name='Feature extraction using LocaNMF to isolate activity from individual sources in the data.',
                                desc = 'Applies locaNMF to the dataset to isolate the activity of individual areas. This step requires a compressed  dataset.',
                                selected = True)]

def build_tree(item,parent):
    for k in item.keys():
        child = QStandardItem(k)
        child.setFlags(child.flags() |
                       Qt.ItemIsSelectable |
                       Qt.ItemIsEnabled)
        child.setEditable(False)
        if type(item[k]) is dict:
            build_tree(item[k],child)
        parent.appendRow(child)


class CredentialsManager(QDockWidget):
    def __init__(self,
                 configfile = pjoin(
                     os.path.expanduser('~'),
                     '.wfield','ncaas_config.json')):
        '''
        Logs to NeuroCaas to retrieve the keys and 
allows changing the config parameters and credentials 
from the GUI.
        '''
        super(CredentialsManager,self).__init__()
        self.awsinfo = ncaas_read_aws_keys()
        self.ncaasconfig = ncaas_read_analysis_config(configfile)
        ncaasconfig_json = json.dumps(self.ncaasconfig,
                                      indent=4,
                                      sort_keys=True)
        self.configfile = configfile
        
        mainw = QWidget()
        self.setWidget(mainw)
        layout = QGridLayout()
        mainw.setLayout(layout)
        
        self.setWindowTitle('NeuroCAAS configuration')

        tabwidget = QTabWidget()
        layout.addWidget(tabwidget,0,0)

        advancedwid = QWidget()
        lay = QFormLayout()
        advancedwid.setLayout(lay)

        self.cred_access = QLineEdit(self.awsinfo['access_key'])
        lay.addRow(QLabel('AWS access key'),self.cred_access)
        def cred_access():
            self.awsinfo['access_key'] = self.cred_access.text()
        self.cred_access.textChanged.connect(cred_access)

        self.cred_secret = QLineEdit(self.awsinfo['secret_key'])
        lay.addRow(QLabel('AWS secret key'),self.cred_secret)
        def cred_secret():
            self.awsinfo['secret_key'] = self.cred_secret.text()
        self.cred_secret.textChanged.connect(cred_secret)
        
        self.aws_region = QComboBox()
        for r in awsregions:
            self.aws_region.addItem(r)
        self.aws_region.setCurrentIndex(awsregions.index(self.awsinfo['region']))
        lay.addRow(QLabel('AWS region'),self.aws_region)
        def region_call(value):
            self.awsinfo['region'] = awsregions[value]    
        self.aws_region.currentIndexChanged.connect(region_call)
        
        self.configedit = QPlainTextEdit(ncaasconfig_json)
        lay.addRow(QLabel('NCAAS settings'),self.configedit)
        w = QWidget()
        l = QVBoxLayout()
        w.setLayout(l)
        lab = QLabel('Log to neurocaas.org and go to the user settings page to get the credentials.')
        l.addWidget(lab)
        web = QWebEngineView()
        l.addWidget(web)
        web.load(QtCore.QUrl("http://www.neurocaas.org/profile/"))
        tabwidget.addTab(w,'NeuroCAAS login')
        tabwidget.addTab(advancedwid,'Settings')

        self.html = ''
        def parsehtml():
            page = web.page()
            def call(var):
                self.html = var
                tt = var.split('\n')
                values = []
                extra = '<input class="form-control" type="text" value="'
                for i,t in enumerate(tt):
                    if extra in t:
                        values.append(t.replace(extra,'').strip(' ').split('"')[0])

                if len(values)>=4:
                    self.cred_access.setText(values[2])
                    self.cred_secret.setText(values[3])
                    self.ncaasconfig['user_group_name'] = values[-1]
                    ncaasconfig_json = json.dumps(self.ncaasconfig,
                                                  indent=4,
                                                  sort_keys=True)
                    self.configedit.setPlainText(ncaasconfig_json)
                    print('Got credentials from the website, you can close this window.')
                    dlg = QDialog()
                    dlg.setWindowTitle('Good job!')
                    but = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
                    def ok():
                        self.close()
                        dlg.accept()
                    but.accepted.connect(ok)
                    but.rejected.connect(dlg.accept)
                    l = QVBoxLayout()
                    lab = QLabel('Got the credentials from the website, you can now close this window to continue. Or adjust defaults in the Advanced tab')
                    lab.setStyleSheet("font: bold")
                    l.addWidget(lab)
                    l.addWidget(but)
                    dlg.setLayout(l)
                    dlg.exec_()
                    
            page.toHtml(call)
        web.loadFinished.connect(parsehtml)
        #self.getsite = QPushButton('Get credentials from website')
        def setHtml(self,html):
            self.html = html
        #def getsite():
            
        #lay.addRow(self.getsite)
        #self.getsite.setStyleSheet("font: bold")
        #self.getsite.clicked.connect(getsite)
        self.show()
    def closeEvent(self,event):
        ncaas_set_aws_keys(**self.awsinfo)
        print('Saved AWS keys.')
        try:
            from io import StringIO
            pars = json.load(StringIO(self.configedit.toPlainText()))
        except Exception as E:
            print('Error in the configuration file, did not save')
            return
        with open(self.configfile,'w') as fd:
            json.dump(pars,fd,indent=4,sort_keys = True)
        event.accept()
        
class TextEditor(QDockWidget):
    def __init__(self,path,
                 s3=None,
                 bucket=None,
                 parent = None,
                 refresh_interval = 2,
                 watch_file = True):
        super(TextEditor,self).__init__()
        self.parent = parent
        self.s3 = s3
        print(bucket)
        self.bucketname = bucket
        self.path = path
        self.original = self.refresh_original()
        mainw = QWidget()
        self.setWidget(mainw)
        lay = QVBoxLayout()
        mainw.setLayout(lay)
        self.tx = QTextEdit()
        self.setWindowTitle('NeuroCAAS file - {0}'.format(self.path))
        lay.addWidget(self.tx)
        self.tx.setText(self.original)
        w = QWidget()
        hl = QFormLayout()
        w.setLayout(hl)
        ckbox = QCheckBox()
        hl.addRow(QLabel('Watch file'),ckbox)

        self.timer = QTimer()
        ckbox.setChecked(watch_file)
        def update():
            ori = self.refresh_original()
            if not ori == self.original:
                self.original = ori
                self.tx.setText(self.original)
        self.timer.timeout.connect(update)
        if watch_file:
            self.timer.start(2000)
        
        def watch(val):
            if val:
                self.timer.start(2000)
            else:
                self.timer.stop()
                print('Timer stopped.')
        ckbox.setChecked(watch_file)
        ckbox.stateChanged.connect(watch)
        lay.addWidget(w)
        self.resize(self.width()*1.2, self.height())
    def refresh_original(self):
        if not self.s3 is None:
            #bucket = self.s3.Bucket(self.bucketname)
            print(self.bucketname,self.path,tempfile)
            try:
                self.s3.download_file(self.bucketname,self.path,tempfile)
                with open(tempfile,'r') as f:
                    return f.read()
            except Exception as err:
                print("File {0} may have been deleted or you have no permission.".format(self.path))
                print(err)
                self.close()
    def closeEvent(self,evt):
        try:
            self.timer.stop()
        except:
            pass
        evt.accept()

class  AnalysisSelectionWidget(QDialog):
    def __init__(self, path, config):
        '''
        Select analysis from to be ran on NCAAS.

        '''
        super(AnalysisSelectionWidget,self).__init__()
        import copy

        self.config = copy.deepcopy(config)

        # there are 2 parameter sets, one for locaNMF and one for PMD
        
        layout = QGridLayout()
        self.setLayout(layout)
        
        self.setWindowTitle('NeuroCAAS job parameters')
        
        tabwidget = QTabWidget()
        layout.addWidget(tabwidget,0,0)

        selectwid = QWidget()
        lay = QFormLayout()
        selectwid.setLayout(lay)
        
        self.selections = analysis_selection_dict
        def ckchanged(ind,chk):
            val = self.selections[ind]['selected'] = chk.isChecked()
            if ind == 1:
                self.selections[2]['checkbox'].setChecked(False)
                self.selections[2]['selected'] = False
                print('Hemodynamics correction needs a compressed dataset.')
        for i,k in enumerate(self.selections):
            self.selections[i]['checkbox'] = QCheckBox()
            name = QLabel(k['name'])
            lay.addRow(self.selections[i]['checkbox'],name)
            self.selections[i]['checkbox'].clicked.connect(partial(ckchanged, i,
                                                                   self.selections[i]['checkbox']))
            self.selections[i]['checkbox'].setToolTip(self.selections[i]['desc'])
            name.setToolTip(self.selections[i]['desc'])
            self.selections[i]['checkbox'].setChecked(self.selections[i]['selected'])
        tabwidget.addTab(selectwid,'Select analysis')

        parameterswid = QWidget()
        lay = QFormLayout()
        parameterswid.setLayout(lay)
        def vchanged(analysis,option_name,edit):
            dtype = type(self.config[analysis]['config'][option_name])
            try:
                self.config[analysis]['config'][option_name] = dtype(edit.text())
                if option_name ==  'num_channels':
                    if not self.config[analysis]['config'][option_name] == 2:
                        #disable hemodynamics
                        self.selections[2]['checkbox'].setChecked(False)
                        self.selections[2]['checkbox'].setEnabled(False)
                        self.selections[2]['selected'] = False
                        print('WARNING: Disabled hemodynamics correction because of num_channels')
                    else:
                        self.selections[2]['checkbox'].setChecked(True)
                        self.selections[2]['selected'] = True
                        self.selections[2]['checkbox'].setEnabled(True)
                        print('WARNING: Enabled hemodynamics correction because of num_channels = 2')
            except:
                print('Could not get {0}.'.format(option_name))
        if PMD_BUCKET in self.config.keys():
            if 'config' in self.config[PMD_BUCKET].keys():
                lay.addRow(QLabel('Parameters for motion correction, compression, denoising and hemodynamics compensation'))

                for k in self.config[PMD_BUCKET]['config'].keys():
                    if not 'analysis_selection' == k:
                        c = QLineEdit(str(self.config[PMD_BUCKET]['config'][k]))
                        name = QLabel(k)
                        lay.addRow(name,c)
                        c.textChanged.connect(partial(vchanged,PMD_BUCKET ,k,c))
                    
        if NMF_BUCKET in self.config.keys():
            if 'config' in self.config[NMF_BUCKET].keys():
                lay.addRow(QLabel('Parameters for feature extraction'))                                  
                for k in self.config[NMF_BUCKET]['config'].keys():
                    c = QLineEdit(str(self.config[NMF_BUCKET]['config'][k]))
                    name = QLabel(k)
                    lay.addRow(name,c)
                    c.textChanged.connect(partial(vchanged,NMF_BUCKET,k,c))
        tabwidget.addTab(parameterswid,'Parameters')

        files = glob(pjoin(path,'*'))
        fileswid = QWidget()
        lay = QFormLayout()
        fileswid.setLayout(lay)
        
        def filechanged(f,chk):
            val = self.files[f]['selected'] = chk.isChecked()
        files = [dict(fname = f,selected=True) for f in files]
        self.files = []
        for i,fname in enumerate(files):
            if not os.path.isfile(fname['fname']):
                continue
            _,ext = os.path.splitext(fname['fname'])
            if not ext in analysis_extensions or 'landmarks.json' in fname['fname']:
                fname['selected'] = False                
            self.files.append(fname)
            f = QCheckBox()
            name = QLabel(os.path.basename(fname['fname']))
            lay.addRow(f,name)
            f.clicked.connect(partial(filechanged, len(self.files)-1,f))
            f.setChecked(fname['selected'])
        if not len(self.files):
            print('This folder {0} contained no files that can be analysed.'.format(path))
        scroll = QScrollArea()
        scroll.setWidget(fileswid)
        tabwidget.addTab(scroll,'File selection')

        advancedwid = QWidget()
        lay = QFormLayout()
        advancedwid.setLayout(lay)
        def vschanged(analysis,option_name,edit):
            dtype = type(self.config[analysis]['submit'][option_name])
            try:
                self.config[analysis]['submit'][option_name] = dtype(edit.text())
            except:
                print('Could not get {0}'.format(option_name))

        if PMD_BUCKET in self.config.keys():
            if 'submit' in self.config[PMD_BUCKET].keys():
                lay.addRow(QLabel('Advanced parameters for motion correction, compression, denoising and hemodynamics compensation'))
                for k in self.config[PMD_BUCKET]['submit'].keys():
                    c = QLineEdit(str(self.config[PMD_BUCKET]['submit'][k]))
                    name = QLabel(k)
                    lay.addRow(name,c)
                    c.textChanged.connect(partial(vschanged,PMD_BUCKET ,k,c))
        if NMF_BUCKET in self.config.keys():
            if 'submit' in self.config[NMF_BUCKET].keys():
                lay.addRow(QLabel('Advanced parameters for feature extraction'))                                  
                for k in self.config[NMF_BUCKET]['submit'].keys():
                    c = QLineEdit(str(self.config[NMF_BUCKET]['submit'][k]))
                    name = QLabel(k)
                    lay.addRow(name,c)
                    c.textChanged.connect(partial(vschanged,NMF_BUCKET,k,c))
        tabwidget.addTab(advancedwid,'Advanced configuration')

        
        submit_button = QPushButton('Submit analysis')
        layout.addWidget(submit_button,1,0)
        self.transfer_queue = []
        self.transfer_config = []
        def submit():
            filetransfer = []
            aws_transfer_queue = []
            for f in self.files:
                if f['selected']:
                    filetransfer.append(f['fname'])
            foldername = os.path.basename(path)
            if len(filetransfer):
                aws_transfer_queue.append(
                    dict(name = foldername, #os.path.basename(filetransfer[0]),
                         awsdestination = ['{0}' + foldername + '{1}'
                                           + os.path.basename(f) for f in filetransfer],
                         awssubmit = '{0}'+foldername.replace('/inputs','/configs')+'/{1}submit.json',
                         awsbucket = None,
                         localpath = filetransfer,
                         log=None,
                         last_status = 'pending_transfer',
                         last_change_time = datetime.now().strftime("%m/%d/%Y, %H:%M:%S")))
            if len(aws_transfer_queue):
                self.transfer_queue = aws_transfer_queue
                self.transfer_config = dict(self.config, selection=self.selections)
                if self.selections[3]['selected']:
                    # Then locaNMF is selected.
                    lmarkfiles = glob(pjoin(path,'*_landmarks.json'))
                    if not len(lmarkfiles):
                        print('Opening the dataset to compute get Allen Landmarks.')
                        from .io import load_stack
                        from .widgets import AllenMatchWidget
                        dat = load_stack(path,imager_preview = True)
                        dlg = QDialog()
                        dlg.setWindowTitle('Allen alignment is required for locaNMF')
                        l = QVBoxLayout()
                        lab = QLabel('Move the point to the respective landmarks and close this window.')
                        lab.setStyleSheet("font: bold")
                        l.addWidget(lab)
                        w = AllenMatchWidget(raw = dat,
                                             folder = path,
                                             mainwidget = dlg)
                        l.addWidget(w)
                        dlg.setLayout(l)
                        dlg.exec_()
                    else:
                        print('Found a landmarks file.')
                    print(path)
            else:
                print('No files selected.')
            self.close()

        submit_button.clicked.connect(submit)    
        self.show()
        
    def closeEvent(self,event):
        event.accept()


def ncaas_upload_queue(path,config = None):
    # Search and upload dat file, respect folder structure, return path
    if sys.platform in ['win32']: # path fix on windows
        if path[0] == '/':
            path = path[1:]
    if not os.path.isdir(path):
        path = os.path.dirname(path)
    print('Selected folder: {0}'.format(path))
    print('Choose the analysis to run.')
    dlg = AnalysisSelectionWidget(path,config)
    dlg.exec_()
    return (dlg.transfer_queue,dlg.transfer_config)
    
class NCAASwrapper(QMainWindow):
    def __init__(self,folder = '.',
                 config = pjoin(os.path.expanduser('~'),
                                '.wfield','ncaas_config.json')):
        super(NCAASwrapper,self).__init__()
        self.setWindowTitle('NeuroCAAS preprocessing')
        folder = os.path.abspath(folder)
        if not config is dict: # then it is a filepath
            config = ncaas_read_analysis_config(config)
        self.config = config
        self.folder = folder
        self.uploading = False
        self.fetching_results = False
        
        self.delete_inputs=True
        self.delete_results=True
        
        awskeys = ncaas_read_aws_keys()
        if awskeys['access_key'] == '' or awskeys['secret_key'] == '':
            self.cred = CredentialsManager()
            self.cred.show()
        
        mainw = QWidget()
        self.setCentralWidget(mainw)
        lay = QHBoxLayout()
        mainw.setLayout(lay)
        
        # Filesystem browser
        self.fs_view = FilesystemView(folder, parent=self)
        self.fs_view.expandToDepth(2)
        # Add the widget with label
        w = QWidget()
        l = QFormLayout()
        w.setLayout(l)
        self.folder = QPushButton(folder)
        #self.folder.setStyleSheet("font: bold")
        lab = QLabel('Select local folder:')
        l.addRow(lab, self.folder)
        lab.setStyleSheet("font: bold")
        lab = QLabel('Local folder view - drag to remote:')
        l.addRow(lab)
        lab.setStyleSheet("font: bold")
        ww = QWidget()
        ll = QVBoxLayout()
        ww.setLayout(ll)
        ll.addWidget(w)
        ll.addWidget(self.fs_view)
        lay.addWidget(ww)
        
        def set_folder():
            self.fs_view.query_root()
        self.folder.clicked.connect(set_folder)
        self.open_editors = []        
        # AWS browser
        self.aws_view = AWSView(config, parent=self)
        self.aws_view.expandToDepth(2)

        # Add the widget with label
        historypath = pjoin(os.path.expanduser('~'),
                            '.wfield','ncaas_gui_log.txt')
        self.historyfile = open(historypath,'a+')
        w = QWidget()
        l = QVBoxLayout()
        w.setLayout(l)
        l.addWidget(QLabel('<b>' + 'NeuroCAAS - {0}'.format(', '.join(self.aws_view.bucketnames) + ' - drop below to upload to the cloud <\b>')))
        l.addWidget(self.aws_view)
        lay.addWidget(w)

        # Update the tranfer q and the log
        self.transferqpath = pjoin(os.path.expanduser('~'),
                            '.wfield','ncaas_transfer_q.json')
        if os.path.isfile(self.transferqpath):
            with open(self.transferqpath,'r') as f:
                try:
                    self.aws_view.aws_transfer_queue = json.load(f)
                except:
                    print('Corrupt transfer queue file?')
        for i,t in enumerate(self.aws_view.aws_transfer_queue):
            if t['last_status'] == 'in_transfer': # There can be no transfers in init - resume
                self.aws_view.aws_transfer_queue[i]['last_status'] = 'pending_transfer'
            if t['last_status'] == 'fetching_results': # there can be no fetching when init - resume
                self.aws_view.aws_transfer_queue[i]['last_status'] = 'submitted'
            t['log'] = None
        
        sw = QWidget()
        sl = QVBoxLayout()
        sw.setLayout(sl)

        self.queuelist = QListWidget()
        self.pbar = QProgressBar()

        self.infomon = QTextEdit()
        self.infomon.insertPlainText(''.join(tail(historypath)))

        self.submitb = QPushButton('Submit to NeuroCAAS')
        self.submitb.setStyleSheet("font: bold")
        
        lab = QLabel('History and neurocaas information:')
        lab.setStyleSheet("font: bold")
        sl.addWidget(lab)
        sl.addWidget(self.infomon)
        lab = QLabel('Job submission queue:')
        lab.setStyleSheet("font: bold")
        sl.addWidget(lab)
        sl.addWidget(self.queuelist)
        sl.addWidget(self.submitb)
        lab = QLabel('Transfer progress:')
        lab.setStyleSheet("font: bold")
        sl.addWidget(lab)
        sl.addWidget(self.pbar)
        l.addWidget(sw)

        self.queuelist.itemDoubleClicked.connect(self.remove_from_queue)
        self.refresh_queuelist()
        self.submitb.clicked.connect(self.process_aws_transfer)
        #import ipdb
        #ipdb.set_trace()
        self.resize(QDesktopWidget().availableGeometry().size() * 0.85);        
        self.show()
        self.fetchresultstimer = QTimer()
        self.fetchresultstimer.timeout.connect(self.fetch_results)
        self.fetchresultstimer.start(5000)
        
    def fetch_results(self):
        '''
        Checks if the data are analyzed and copies the results. 
        '''
        if self.uploading:
            return
        
        for i,t in enumerate(self.aws_view.aws_transfer_queue):
            resultsdir = os.path.dirname(t['awsdestination'][0]).replace('/inputs',
                                                                         '/results')
            outputsdir = os.path.dirname(t['awsdestination'][0]).replace('/inputs',
                                                                         '/outputs')
            logsdir = os.path.dirname(t['awsdestination'][0]).replace('/inputs',
                                                                      '/results')
            if t['last_status'] == 'submitted':
                if t['log'] is None: # check of there is any element on the queue that needs a log to open
                    logs = []
                    for a in self.aws_view.awsfiles:
                        if logsdir in a:
                            logs.append(a)
                    if len(logs):
                        t['log'] = logs
                        for l in logs:
                            extension = os.path.splitext(l)[-1]
                            if extension in ['.yaml','.txt','.json']:
                                bucket = l.strip('/').split('/')[0]
                                temp = l.replace(bucket,'').strip('/')
                                wid = TextEditor(temp, s3=self.aws_view.s3client, bucket = bucket)
                                self.addDockWidget(Qt.RightDockWidgetArea, wid)
                                wid.setFloating(True)
                resultsfiles = []
                awsfiles = self.aws_view.awsfiles
                #awsfiles = s3_ls(self.aws_view.s3,self.aws_view.bucketnames) # update directly here (no lag).
                for a in awsfiles:
                    if resultsdir in a or outputsdir in a:
                        resultsfiles.append(a)
                if len(resultsfiles):
                    # check that the results are all there (this should change in the next implementation of neurocaas, hopefully)
                    if t['awsbucket'] == PMD_BUCKET:
                        got_all_results = [False,False]
                        for j,res in enumerate(['reduced_spatial.npy',
                                                'reduced_temporal.npy']):
                            for f in resultsfiles:
                                if res in f:
                                    got_all_results[j] = True
                        got_all_results = np.sum(got_all_results) == 2
                    elif t['awsbucket'] == NMF_BUCKET:
                        got_all_results = [False,False]
                        for j,res in enumerate(['SummaryComponents.png',
                                                'LocaNMF_Components.mat']):
                            for f in resultsfiles:
                                if res in f:
                                    got_all_results[j] = True
                        got_all_results = np.sum(got_all_results) == 2
                    else:
                        got_all_results = True # Don't bother doing this for locaNMF?
                    if got_all_results == False:
                        #print('Not all files were there for the PMD bucket?')
                        return # Do nothing
                    self.submitb.setEnabled(False)
                    self.submitb.setText('Downloading result files')                        

                    self.aws_view.aws_transfer_queue[i]['last_status'] = 'fetching_results'
                    self.aws_view.aws_transfer_queue[i]['last_change_time'] = datetime.now().strftime("%m/%d/%Y, %H:%M:%S")
                    self.refresh_queuelist()
                    for a in self.aws_view.awsfiles:
                        if logsdir in a:
                            resultsfiles.append(a)
                    print('Found results for {name}'.format(**t))
                    if not os.path.dirname(t['localpath'][0]).endswith('results'):
                        localpath = pjoin(os.path.dirname(t['localpath'][0]),'results')
                    else:
                        localpath = os.path.dirname(t['localpath'][0])
                    localpath = os.path.abspath(localpath)
                    if not os.path.exists(localpath):
                        os.makedirs(localpath)
                        self.to_log('Creating {0}'.format(localpath))
                    bucket = self.aws_view.s3.Bucket(t['awsbucket'])
                    self.fetching_results = True
                    safe_delete = True
                    t['downloaded'] = []
                    t['safe_delete'] = True
                    def get():
                        for f in resultsfiles:
                            # drop the bucket name
                            fn = f.replace(t['awsbucket']+'/','')
                            if outputsdir in f:
                                lf = fn.replace(outputsdir,'{localpath}/')
                            if resultsdir in f:
                                lf = fn.replace(resultsdir,'{localpath}/')
                            if logsdir in f:
                                lf = fn.replace(logsdir,'{localpath}/')
                            lf = lf.split('/')
                            lf = pjoin(*lf).format(localpath = localpath)
                            if not os.path.isdir(os.path.dirname(lf)):
                                os.makedirs(os.path.dirname(lf))
                            print(fn,lf)
                            t['downloaded'].append(lf)
                            if not os.path.isdir(lf):
                                try:
                                    bucket.download_file(fn,lf)
                                except:
                                    print('Warning: Could not download the file {0}. Is it open?'.format(lf))
                                    t['safe_delete'] = False
                                    
                            self.to_log('Fetching {0}'.format(lf))
                    safe_delete=t['safe_delete']
                    thread = threading.Thread(target=get)
                    thread.start()
                    time.sleep(0.5)
                    while thread.is_alive():
                        QApplication.processEvents()
                        time.sleep(0.02)
                    self.aws_view.aws_transfer_queue[i]['last_status'] = 'got_results'
                    
                    if not safe_delete:
                        t['last_status'] = 'submitted'
                        t['last_change_time'] = datetime.now().strftime("%m/%d/%Y, %H:%M:%S")

                        self.to_log('Could not fetch results.')
                        self.submitb.setEnabled(True)
                        self.submitb.setText('Submit to NeuroCAAS')
                        return

                    if self.delete_results:
                        for f in resultsfiles:
                            self.to_log('Remote delete: {0}'.format(f))
                            f = f.replace(t['awsbucket']+'/','')
                            try:
                                self.aws_view.s3.Object(t['awsbucket'],f).delete()
                            except: # delete already happened?
                                pass
                    if self.delete_inputs:
                        # need to delete the remote data
                        self.to_log('Remote delete: {0}'.format(t['awsdestination']))
                        configpath = t['awssubmit'].replace('submit.json','config.yaml')
                            #os.path.dirname(t['awsdestination'][0])+'/'+'config.yaml' 
                        submitpath = t['awssubmit']#os.path.dirname(t['awsdestination'])+'/'+'submit.json'
                        
                        for a in t['awsdestination']:
                            try:
                                self.aws_view.s3.Object(t['awsbucket'],a).delete()
                            except:
                                pass
                            try:
                                self.aws_view.s3.Object(t['awsbucket'],configpath).delete()
                                self.aws_view.s3.Object(t['awsbucket'],submitpath).delete()
                            except:
                                pass
                    self.to_log('Done fetching results to {0}'.format(localpath))
                    self.to_log('COMPLETED {0}'.format(t['name']))
                
                    if 'analysis_selection' in t['config'].keys():
                        if not 'locanmf' in t['awsbucket']:
                            if 'locaNMF' in t['config']['analysis_selection']:
                                print('Got results, submitting locaNMF analysis')
                                nt = dict(**t)
                                nt['name'] = nt['name'] + '_locanmf'
                                awsfolder = os.path.dirname(nt['awsdestination'][0])
                                foldername = awsfolder.strip('/').split('/')[-1]
                                if not 'locanmf_config' in t.keys():
                                    print('''
There was no locanmf config in the submittion job. 

This happens when you re-submit. You need to resubmit from uploaded data.''')
                                    break
                                nt['config'] = t['locanmf_config']
                                nt['submit'] = t['locanmf_submit']
                                nt['submit']['userfolder'] = self.config['user_group_name']
                                nt['awsbucket'] = NMF_BUCKET
                                # LocalNMF is going to upload only to configs
                                nt['awssubmit'] = ('{0}/{1}/'+foldername+'/submit.json').format(
                                    nt['submit']['userfolder'],'configs')
                                #print(nt['awssubmit'])
                                # Which files: labels.json,atlas.npy,brainmask.npy,SVTcorr.npy,U.npy
                                #t['awsdestination'][i] = f.format(t['submit']['userfolder'],'/inputs/')
                                # Check for the landmarks file in the original folder
                                localfolder = os.path.dirname(nt['localpath'][0])
                                landmarksfile = glob(pjoin(localfolder,'*_landmarks.json'))
                                if len(landmarksfile):
                                    landmarksfile = landmarksfile[0]
                                else:
                                    landmarksfile = None
                                if landmarksfile is None:
                                    print('Need to do the alignment first using ncaas open_raw')
                                for d in t['downloaded']:
                                    if 'reduced_spatial' in d:
                                        print('Found reduced spatial.')
                                        U = np.load(d)
                                        # This is in [H,W,k] size
                                    #if 'config.yaml' in d:
                                    #    with open(d,'r') as fd:
                                    #        import yaml
                                    #        config = yaml.load(fd)
                                    #        dims = [config['fov_height'],config['fov_width']]
                                
                                if  'U' in dir() and not landmarksfile is None:
                                    print('Creating the atlas.')
                                    dims = U.shape[:-1]
                                    from .allen import atlas_from_landmarks_file, load_allen_landmarks
                                    atlas, areanames, brain_mask = atlas_from_landmarks_file(landmarksfile,
                                                                                             dims = dims,
                                                                                             do_transform = False)
                                    nt['localpath'] = []
                                    nt['awsdestination'] = []
                                    print('Warping the U matrix.')
                                    from .utils import runpar, im_apply_transform
                                    U[:,0,:] = 1e-10
                                    U[0,:,:] = 1e-10
                                    U[-1,:,:] = 1e-10
                                    U[:,-1,:] = 1e-10
                                    lmarks = load_allen_landmarks(landmarksfile)
                                    try:
                                        U = np.stack(runpar(
                                            im_apply_transform,
                                            U.transpose([2,0,1]),
                                            M = lmarks['transform'])).transpose([1,2,0])
                                    except:
                                        print("There was an error trying to transpose the spatial components..")
                                        break
                                    # spatial
                                    fname = pjoin(localfolder,'results','U_atlas.npy')
                                    np.save(fname,U)
                                    nt['submit']['spatial_data_filename'] = os.path.basename(fname)
                                    f = '{0}/' + foldername + '{1}'+'U_atlas.npy'
                                    nt['localpath'].append(fname)
                                    nt['awsdestination'].append(f.format(nt['submit']['userfolder']+'/configs','/inputs/'))
                                    # temporal
                                    if os.path.isfile(pjoin(localfolder,'results','SVTcorr.npy')):
                                        fname = pjoin(localfolder,'results','SVTcorr.npy')
                                    elif os.path.isfile(pjoin(localfolder,'results','reduced_temporal.npy')):
                                        fname = pjoin(localfolder,'results','reduced_temporal.npy')
                                        print('Using reduced temporal for locaNMF temporal!!')
                                    else:
                                        print('No temporal components?')
                                        break
                                    nt['submit']['temporal_data_filename'] = os.path.basename(fname)
                                    f = '{0}/' + foldername + '{1}'+os.path.basename(fname)
                                    nt['localpath'].append(fname)
                                    nt['awsdestination'].append(f.format(nt['submit']['userfolder']+'/configs','/inputs/'))

                                    # brainmask
                                    fname = pjoin(localfolder,'results','labels.json')
                                    areanames = [dict(acronym = t[1],
                                                      label=t[0]) for t in areanames]
                                    with open(fname,'w+') as fd:
                                        json.dump(areanames, fd, indent=4)
                                    nt['submit']['areanames_filename'] = os.path.basename(fname)
                                    f = '{0}/' + foldername + '{1}'+'labels.json'
                                    nt['localpath'].append(fname)
                                    nt['awsdestination'].append(f.format(nt['submit']['userfolder']+'/configs','/inputs/'))
                                    
                                    # brainmask
                                    fname = pjoin(localfolder,'results','brainmask.npy')
                                    np.save(fname,brain_mask)
                                    nt['submit']['brainmask_filename'] = os.path.basename(fname)
                                    f = '{0}/' + foldername + '{1}'+'brainmask.npy'
                                    nt['localpath'].append(fname)
                                    nt['awsdestination'].append(f.format(nt['submit']['userfolder']+'/configs','/inputs/'))

                                    #atlas
                                    fname = pjoin(localfolder,'results','atlas.npy')
                                    np.save(fname,atlas)
                                    nt['submit']['atlas_filename'] = os.path.basename(fname)
                                    f = '{0}/' + foldername + '{1}'+'atlas.npy'
                                    nt['localpath'].append(fname)
                                    nt['awsdestination'].append(f.format(nt['submit']['userfolder']+'/configs','/inputs/'))
                                    
                                    nt['last_status'] = 'pending_transfer'
                                    nt['last_change_time'] = datetime.now().strftime("%m/%d/%Y, %H:%M:%S")
                                    
                                    self.aws_view.aws_transfer_queue.append(nt)
                                    #print(nt)
                                    self.refresh_queuelist()
                                    while self.uploading:
                                        QApplication.processEvents()
                                        time.sleep(0.02)
                                    self.process_aws_transfer()
                                    self.aws_view.aws_transfer_queue[-1]['last_status'] = 'submitted'
                                    self.aws_view.aws_transfer_queue[-1]['last_change_time'] = datetime.now().strftime("%m/%d/%Y, %H:%M:%S")
                    self.fetching_results = False
                    #self.remove_from_queue(self.queuelist.item(i))
                    self.refresh_queuelist()
                    self.submitb.setEnabled(True)
                    self.submitb.setText('Submit to NeuroCAAS')
                    return # because we removed an item from the queue, restart the loop
                

    def remove_from_queue(self,item):
        if not self.uploading:
            itemname = item.text()
            names = [t['name'] for t in self.aws_view.aws_transfer_queue]
            #itemname = itemname..strip('Dataset ')split('-')[0].strip(' ')
            ii = None
            for i in range(len(names)):
                if names[i] in itemname:
                    ii = i
                    break
            #ii = names.index(itemname)
            tt = self.aws_view.aws_transfer_queue.pop(ii)
            self.to_log('Removed {name} from the tranfer queue'.format(**tt))
            self.queuelist.takeItem(self.queuelist.row(item))
            self.refresh_queuelist()
        else:
            print('Wait until the upload finishes to remove from queue.')
            return
    def to_log(self,msg):
        msg  = to_log(msg,logfile = self.historyfile)
        self.infomon.moveCursor(-1)
        self.infomon.insertPlainText(msg)

    def store_transfer_queue(self):
        with open(self.transferqpath,'w') as fp:
            json.dump(self.aws_view.aws_transfer_queue,fp, indent=4)

    def refresh_queuelist(self,store=True):
        if store:
            self.store_transfer_queue()
        for i,itt in enumerate(self.aws_view.aws_transfer_queue):
            if i >= self.queuelist.count():
                it = QListWidgetItem(itt['name'])
                self.queuelist.insertItem(self.queuelist.count(),it)
            item = self.queuelist.item(i)
            if not 'last_change_time' in itt.keys():
                itt['last_change_time'] = datetime.now().strftime("%m/%d/%Y, %H:%M:%S")
            item.setText('Dataset {0} - state {1} [{2}]'.format(
                itt['name'],
                itt['last_status'].replace('_',' '),
                itt['last_change_time']))
            if itt['last_status'] == 'pending_transfer':
                item.setForeground(QColor(176,107,12))
            elif itt['last_status'] in ['in_transfer','fetching_results']:
                item.setForeground(QColor(7,106,139))
            elif itt['last_status'] in ['uploaded']:
                item.setForeground(QColor(72,33,17))
            elif itt['last_status'] == 'submitted':
                item.setForeground(QColor(22,66,34))
            else:
                item.setForeground(QColor(0,0,0))
                

    def process_aws_transfer(self):
        if self.uploading:
            print('Upload in progress')
            return
        self.uploading = True
        self.submitb.setEnabled(False)
        self.submitb.setText('Transfer in progress')                        
        QApplication.processEvents()
        for i,t in enumerate(self.aws_view.aws_transfer_queue):
            if t['last_status'] == 'pending_transfer':
                # Check if files need to be zipped.
                if t['awsbucket'] == PMD_BUCKET :#and len(t['awsdestination']) > 1: # this should change in the future
                    docompress = [True]
                    from .utils import zipfiles
                    dname = os.path.dirname(t['localpath'][0])
                    fname = t['name']
                    fname = pjoin(dname, fname + '.zip')
                    if os.path.isfile(fname):
                        dlg = QDialog()
                        dlg.setWindowTitle('Found a zip file, do you upload that file?')
                        but = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
                        def ok():
                            docompress[0] = False
                            dlg.accept()
                        but.accepted.connect(ok)
                        but.rejected.connect(dlg.accept)
                        l = QVBoxLayout()
                        lab = QLabel('Found a zip file, do you upload that file?')
                        lab.setStyleSheet("font: bold")
                        l.addWidget(lab)
                        l.addWidget(but)
                        dlg.setLayout(l)
                        dlg.exec_()
                    if docompress[0]:
                        self.to_log('Zipping {0}'.format(fname))
                        self.submitb.setText('Compressing files')                        
                        QApplication.processEvents()
                        zipfiles(t['localpath'], fname)
                    localpath = [fname]
                    awsdestination = [os.path.dirname(t['awsdestination'][0]) +
                                      '/'+ os.path.basename(fname)]
                else:
                    localpath = t['localpath']
                    awsdestination = t['awsdestination']
                self.submitb.setText('Transfer in progress')                        
                QApplication.processEvents()
                for it, fname in enumerate(localpath):
                    if os.path.isfile(fname):
                        self.pbar.setValue(0)
                        self.pbar.setMaximum(100)
                        self.aws_view.aws_transfer_queue[i]['last_status'] = 'in_transfer'
                        self.aws_view.aws_transfer_queue[i]['last_change_status'] = datetime.now().strftime("%m/%d/%Y, %H:%M:%S")
                        self.refresh_queuelist()
                        self.count = 0
                        # check if the file is already there
                        dotransfer = [True]
                        for remotef in self.aws_view.awsfiles:
                            if awsdestination[it] in remotef:
                                dotransfer[0] = False
                        if not dotransfer[0]:
                            dlg = QDialog()
                            dlg.setWindowTitle('File upload {0}'.format(
                                os.path.basename(awsdestination[it])))
                            but = QDialogButtonBox(QDialogButtonBox.Yes | QDialogButtonBox.No)
                            def ok():
                                dotransfer[0] = True
                                dlg.accept()
                            but.accepted.connect(ok)
                            but.rejected.connect(dlg.accept)
                            l = QVBoxLayout()
                            lab = QLabel('A similar file is already on the cloud, do you want to replace it?')
                            lab.setStyleSheet("font: bold")
                            l.addWidget(lab)
                            l.addWidget(but)
                            dlg.setLayout(l)
                            dlg.exec_()
                        if not dotransfer[0]:
                            print('Skipped {0}'.format(localpath[it]))
                            continue
                        print(awsdestination[it])
                        upload = Upload(bucket = t['awsbucket'],
                                        filepath = localpath[it],
                                        destination = awsdestination[it],
                                        s3 = self.aws_view.s3)
                        
                        upload.start()
                        time.sleep(.1)
                        cnt = 0
                        while (upload.isrunning):
                            QApplication.processEvents()
                            self.pbar.setValue(np.ceil(upload.count*99/upload.fsize))
                            time.sleep(0.033)
                            cnt+= 1
                            if np.mod(cnt,3) == 0:
                                self.submitb.setStyleSheet("color: red")
                            else:
                                self.submitb.setStyleSheet("color: black")
                        
                        self.to_log('Transfered {0} {1}'.format(localpath[it],
                                                                 awsdestination[it]))
                        if 'docompress' in dir():
                            if docompress[0]:
                                if localpath[it].endswith('.zip'):
                                    self.to_log('Deleting the zip file {0}'.format(localpath[it]))
                                    os.remove(localpath[it])
                        QApplication.processEvents()
                    else:
                        self.to_log('File not found {localpath}'.format(**t))
                        self.remove_from_queue(self.queuelist.item(i))
                        self.submitb.setStyleSheet("color: black")
                        self.submitb.setEnabled(False)
                        return
                    self.to_log('Done transfering {name}'.format(**t))
                t['awsdestination'] = awsdestination
                self.submitb.setStyleSheet("color: black")
                self.pbar.setValue(0)
                self.aws_view.aws_transfer_queue[i]['last_status'] = 'uploaded'
                self.aws_view.aws_transfer_queue[i]['last_change_status'] = datetime.now().strftime("%m/%d/%Y, %H:%M:%S")
                t['last_status'] = 'uploaded'
            if t['last_status'] == 'uploaded':
                # add a config file
                import yaml
                temp = pjoin(os.path.expanduser('~'),'.wfield','temp_config.yaml')
                with open(temp,'w') as f: 
                    yaml.dump(t['config'],f)
                bucket =self.aws_view.s3.Bucket(t['awsbucket'])
                bucket.upload_file(temp,
                                   #t['awssubmit'].replace('submit.json', 'config.yaml'))
                                   os.path.dirname(t['awsdestination'][0])+'/'+'config.yaml')
                self.to_log('Uploaded default config to {name}'.format(**t))
                
                temp = pjoin(os.path.expanduser('~'),'.wfield','temp_submit.json')
                t['submit'] = {k:t['submit'][k] for k in t['submit'] if not k in ['userfolder',
                                                                                  'analysis_extension',
                                                                                  'decompress_results']}
                tmp = dict(t['submit'],
                           dataname = os.path.dirname(t['awsdestination'][0])+'/')
                with open(temp,'w') as f:
                    json.dump(tmp, f, indent=4, sort_keys=True)
                bucket.upload_file(temp,
                                   t['awssubmit'])
                # this is to be able to load the file.
                if not '/configs/' in t['awssubmit']:
                    bucket.upload_file(temp,
                                       t['awssubmit'].replace('/inputs/','/configs/'))

                self.to_log('Submitted analysis {name} to {awssubmit}'.format(**t))
                self.aws_view.aws_transfer_queue[i]['last_status'] = 'submitted'
                self.aws_view.aws_transfer_queue[i]['last_change_time'] = datetime.now().strftime("%m/%d/%Y, %H:%M:%S")
            self.refresh_queuelist()
        self.uploading = False
        self.submitb.setText('Run on NeuroCAAS')                        
        self.submitb.setEnabled(True)

                
        # for f in filetransfer:
        #    fsize = statinfo.st_size

class AWSView(QTreeView):
    def __init__(self,config,parent=None):
        super(AWSView,self).__init__()
        self.parent = parent
        self.setUniformRowHeights(True)
        self.setAlternatingRowColors(True)
        self.setSelectionMode(3)
        self.setAcceptDrops(True)
        self.setDragDropMode(QAbstractItemView.InternalMove)
        self.setDropIndicatorShown(True) 
        self.setAutoScroll(True) 

        # add right click support
        
        def menu_here(event):
            self.menu = QMenu(self)
            delete = self.menu.addAction("Delete")
            ii = self.indexAt(event)
            item = self.model.index(ii.row(),0,ii.parent())
            path = get_tree_path([item])
            if path[-1].endswith('submit.json'):
                rerun = self.menu.addAction("Re-submit")
            tmp = self.menu.exec_(self.mapToGlobal(event))
            if tmp is not None:
                if tmp == delete:
                    for p in path:
                        bucketname = p.strip('/').split('/')[0]
                        temp = p.replace(bucketname,'').strip('/')
                        if not self.s3 is None:
                            self.s3.Object(bucketname,temp).delete()
                            print(bucketname,temp)
                elif tmp == rerun:
                    # Download and upload to re-submit. 
                    if not self.s3 is None:
                        for p in path:
                            bucketname = p.strip('/').split('/')[0]
                            fname = p.replace(bucketname,'').strip('/')
                            
                            bucket = self.s3.Bucket(bucketname)
                            print('This is not possible with this version of neurocaas, it will be available in the future.')
                            return
                            bucket.download_file(fname,tempfile)
                            dicttmp = {}
                            dicttmp['awssubmit'] = fname
                            with open(tempfile,'r') as fd:
                                temp = json.load(fd)
                                dicttmp['submit'] = temp
                            dname = temp['dataname']
                            files = []
                            for f in self.awsfiles: # find the datfile
                                if dname in f:
                                    if 'config.yaml' in f:
                                        bucket.download_file(f.replace(bucketname,'').strip('/'),tempfile)
                                        with open(tempfile,'r') as fd:
                                            dicttmp['config'] = yaml.load(fd)
                                        continue
                                    if 'submit.json' in f:
                                        bucket.download_file(f.replace(bucketname,'').strip('/'),tempfile)
                                        with open(tempfile,'r') as fd:
                                            dicttmp['submit'] = json.load(fd)
                                        continue
                                    files.append(f.replace(bucketname,'').strip('/'))
                                    if 'inputs' in f:
                                        tt = f.split('/')
                                        localpath = None
                                        for i,t in enumerate(tt): # todo: re-write
                                            if t == 'inputs':
                                                localpath = os.path.abspath(pjoin(os.path.curdir,*tt[i+1:-1]))
                                        if localpath is None:
                                            self.to_log('Could not set local folder to re-submit.')
                                        localpath = os.path.curdir
                                    
                            if len(files):
                                toadd = dict(
                                    dicttmp, name = os.path.basename(files[0]),
                                    awsdestination = files,
                                    awsbucket=bucketname,
                                    localpath = [pjoin(localpath,os.path.basename(files[0]))],
                                    last_status = "uploaded",
                                    last_change_time = datetime.now().strftime("%m/%d/%Y, %H:%M:%S"))
                                self.aws_transfer_queue.append(toadd)
                                self.parent.refresh_queuelist()
                                self.parent.to_log('Re-submitted {0} to the queue, (press "Run on NCAAS" to run).'.format(path))
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(menu_here)
        
        self.config = config
        self.user_group_name = config['user_group_name']
        self.s3, self.s3client = s3_connect()
        print('Connected to Amazon Web Services')
        self.bucketnames = self.config.keys()
        self.bucketnames = list(filter( lambda x: not x == 'user_group_name', self.bucketnames))
        print('Using buckets: {0}'.format(', '.join(self.bucketnames)))
        self.awsfiles = []
        self.aws_transfer_queue = []

        self.model = AWSItemModel()
        self.setModel(self.model)
        #aws_model.setEditable(False)
        def open_file(value):
            paths = get_tree_path([value])
            extension = os.path.splitext(paths[0])[-1]
            if extension in ['.yaml','.txt','.json']:
                bucket = paths[0].strip('/').split('/')[0]
                temp = paths[0].replace(bucket,'').strip('/')
                wid = TextEditor(temp, s3=self.s3client, bucket = bucket)
                self.parent.addDockWidget(Qt.RightDockWidgetArea, wid)
                wid.setFloating(True)
                self.parent.open_editors.append(paths[0])
            elif extension == '':
                print('Folder: {0}'.format(paths[0]))
        self.doubleClicked.connect(open_file)
        self.update_files()
        # These cause refresh, need to check if there are new files first.
        self.timer_update = QTimer()
        self.timer_update.timeout.connect(self.update_files)
        self.timer_update.start(4000)

    def update_files(self):
        awsfiles = s3_ls(self.s3client,self.s3,self.bucketnames,folder = self.user_group_name)
        if len(awsfiles) == len(self.awsfiles):
            return
        self.awsfiles = awsfiles
        self.model.clear()
        self.model.setHorizontalHeaderLabels([', '.join(self.bucketnames)])
        root = QStandardItem()
        filetree = {}
        [make_tree(i.split("/"), filetree) for i in self.awsfiles]
        build_tree(filetree,root)
        self.model.appendRow(root)
        #index = self.aws_model.indexFromItem(root)
        #self.expandAll()
    def dragEnterEvent(self, e):        
        if e.mimeData().hasUrls():
            self.setSelectionMode(1)
            e.accept()
        else:
            e.ignore()
    def dragMoveEvent(self, e):
        item = self.indexAt(e.pos())
        self.setCurrentIndex(item)
        e.accept()

    def dragLeaveEvent(self, e):
        self.setSelectionMode(3)
        e.accept()
        
    def dropEvent(self, e):
        paths = get_tree_path([self.indexAt(e.pos())])
        self.setSelectionMode(3)
        [print('Selecting from: {1} to aws {0}'.format(
            paths[0],
            p.path())) for p in e.mimeData().urls()]
        for p in e.mimeData().urls():
            path = p.path()
            transfer, configselection = ncaas_upload_queue(
                path, config = self.config)
            # this is where we parse the bucket.
            if len(configselection) == 0:
                print('Nothing selected')
                continue
            selection = configselection['selection']
            if (selection[0]['selected'] or # motion
                selection[1]['selected'] or # compression
                selection[2]['selected']):  # hemodynamics
                bucketname = PMD_BUCKET 
            elif selection[3]['selected']:
                bucketname = NMF_BUCKET 
            else:
                print('No analysis were selected ? ')
                return e.accept()
            print('Running analysis on {0}'.format(bucketname))
            
            for t in transfer:
                print('Placing {0} in the transfer queue.'.format(t['name']))
                t['awsbucket'] = bucketname
                t['config'] = configselection[bucketname]['config']
                t['config']['analysis_selection'] = [s['acronym'] for s in selection if s['selected']]
                t['submit'] = configselection[bucketname]['submit']
                t['submit']['userfolder'] = self.config['user_group_name']

                if bucketname == PMD_BUCKET:
                    t['awssubmit'] = t['awssubmit'].format(t['submit']['userfolder']+'/inputs/','')
                else:
                    t['awssubmit'] = t['awssubmit'].format(t['submit']['userfolder']+'/configs/','') # goes to configs on locaNMF
                for i,f in enumerate(t['awsdestination']):
                    if bucketname == PMD_BUCKET:
                        t['awsdestination'][i] = f.format(t['submit']['userfolder']+'/inputs/','/')
                    else:
                        t['awsdestination'][i] = f.format(t['submit']['userfolder']+'/inputs/','/')

                # Check if it will run locaNMF
                t['selection'] = [s['acronym'] for s in configselection['selection'] if s['selected']]
                if 'locaNMF' in t['selection']:
                    t['locanmf_config'] = configselection[NMF_BUCKET]['config'] 
                    t['locanmf_submit'] = configselection[NMF_BUCKET]['submit'] 
                added = False
                t['log'] = None
                if len(self.aws_transfer_queue): # check if it is already there
                    names = [a['name'] for a in self.aws_transfer_queue]
                    if not t['name'] in names:
                        added = True
                        self.aws_transfer_queue.append(t)
                    else:
                        self.parent.to_log('{name} was already in the transfer queue'.format(**t))
                else:
                    added = True
                    self.aws_transfer_queue.append(t)
            if added:
                self.parent.to_log('Added {name} to transfer queue'.format(**t))
            self.parent.refresh_queuelist()
        e.ignore() # Dont drop the remote table
        
class AWSItemModel(QStandardItemModel):
    def __init__(self):
        super(AWSItemModel,self).__init__()

    def mimeData(self,idx):
        tt = QMimeData()
        tt.setText(','.join(get_tree_path(idx)))
        return tt
        
class FilesystemView(QTreeView):
    def __init__(self,folder,parent=None):
        super(FilesystemView,self).__init__()
        self.parent = parent
        self.fs_model = QFileSystemModel(self)
        self.fs_model.setReadOnly(True)
        self.setModel(self.fs_model)
        self.folder = folder
        self.setRootIndex(self.fs_model.setRootPath(folder))
        self.fs_model.removeColumn(1)
        self.setAlternatingRowColors(True)
        self.setSelectionMode(3)
        self.setDragEnabled(True)
        self.setAcceptDrops(True)
        self.setDragDropMode(QAbstractItemView.DragDrop)
        self.setDropIndicatorShown(True)
        #[self.hideColumn(i) for i in range(1,4)]
        self.setColumnWidth(0,self.width()*.7)
    def query_root(self):
        folder = QFileDialog().getExistingDirectory(self,"Select directory",os.path.curdir)
        self.setRootIndex(self.fs_model.setRootPath(folder))
        self.expandAll()
        self.folder = folder
        if hasattr(self.parent,'folder'):
            self.parent.folder.setText('{0}'.format(folder))

    def dragEnterEvent(self, e):

        item = self.indexAt(e.pos())
        if e.mimeData().hasText():
            self.setSelectionMode(1)
            e.accept()
        else:
            e.ignore()

    def dragMoveEvent(self, e):
        item = self.indexAt(e.pos())
        self.setCurrentIndex(item)
        e.accept()

    def dragLeaveEvent(self, e):
        self.setSelectionMode(3)
        e.accept()
        
    def dropEvent(self, e):
        idx = self.model().index(
            self.indexAt(e.pos()).row(),
            0,self.indexAt(e.pos()).parent())
        paths = get_tree_path([idx])
        to_fetch = e.mimeData().text()
        
        files = []
        for a in self.parent.aws_view.awsfiles:
            if to_fetch.strip('/') in a:
                files.append(a)
        if not len(files):
            print('No files listed.')
            e.ignore()
            return
        path = [pjoin(self.folder,os.path.basename(p)) for p in paths][0]
        path = os.path.abspath(path)
        if os.path.isfile(path):
            path = os.path.dirname(path)
        if len(files)>1:
            # Then there are multiple files, it must be a folder
            localpath = pjoin(path,os.path.basename(to_fetch))
        else:
            localpath = path
        if not os.path.isdir(localpath):
            os.makedirs(localpath)
        self.parent.to_log('Fetching to {0}'.format(localpath))

        for f in files:
            bucketname = f.strip('/').split('/')[0]
            fn = f.replace(bucketname,'').strip('/')
            bucket = self.parent.aws_view.s3.Bucket(bucketname)
            def get():
                t = pjoin(f.replace(to_fetch.strip('/'),str(localpath)))
                if not os.path.isdir(os.path.dirname(t)):
                    if os.makedirs(os.path.dirname(t)):
                        os.makedirs(os.path.dirname(t))
                bucket.download_file(fn,t)
            thread = threading.Thread(target=get)
            thread.start()
            self.parent.to_log('MANUAL COPY {0}'.format(f))
            while thread.is_alive():
                QApplication.processEvents()
                time.sleep(0.1)
        self.parent.to_log('Done fetching results to {0}'.format(localpath))
        #if self.parent.config['decompress_results']:
            #try:
            #    for f in files:
                    # read U and decompress
                    # this should become a function 
                    #if 'sparse_spatial.npz' in f:
                    #    fname = pjoin(localpath,'sparse_spatial.npz')
                    #    fcfg =  pjoin(localpath,'config.yaml')
                    #    if os.path.isfile(fcfg):
                    #        with open(fcfg,'r') as fc:
                    #            config = yaml.load(fc)
                    #            H,W = (config['fov_height'],config['fov_width'])
                    #        if os.path.isfile(fname):
                    #            from scipy.sparse import load_npz
                    #            Us = load_npz(fname)
                    #            U = np.squeeze(np.asarray(Us.todense()))
                    #            U = U.reshape([H,W,-1])
                    #            np.save(fname.replace('sparse_spatial.npz','U.npy'),U)
                    #            self.parent.to_log('Decompressed {0}'.format(f))
                    #        else:
                    #            print('Could not decompress (no file)')
                    #    else:
                    #        print('Could not decompress (no config.yaml?)')
            #except Exception as err:
            #    self.parent.to_log('ERROR: FAILED TO DECOMPRESS. The error was dumped to the console.')
            #    print(err)
        msg = QMessageBox()
        msg.setIcon(QMessageBox.Critical)
        msg.setText("Do you want to delete the remote files?")
        msg.setInformativeText("Do you want to delete remote files:")
        msg.setWindowTitle("[NCAAS] Remote delete?")
        msg.setDetailedText("The following files will be deleted: "+'\n'.join(files))
        msg.setStandardButtons(QMessageBox.Ok | QMessageBox.No)
        delete_files = dict(msg = False)
        def msgboxok(b):
            if 'OK' in b.text():
                delete_files['msg'] = True
            else:
                print(b.text())
        msg.buttonClicked.connect(msgboxok)
        msg.exec_()
        delete_files = delete_files['msg']

        if delete_files:
            # need to delete the remote data
            for f in files:
                f = f.replace(bucketname,'').strip('/')
                try:
                    self.parent.aws_view.s3.Object(bucketname,f).delete()
                except:
                    # If it was already deleted.
                    pass
                self.parent.to_log('Remote delete: {0}'.format(f))

        self.parent.to_log('FINISHED MANUAL COPY {0}'.format(to_fetch))
        self.setSelectionMode(3)
        e.ignore()
    

def main(folder = '.'):
    configpath = pjoin(os.path.expanduser('~'),'.wfield','ncaas_config.json')
    if QApplication.instance() != None:
        app = QApplication.instance()
    else:
        app = QApplication(sys.argv)
    awskeys = ncaas_read_aws_keys()
    ncaasconfig = ncaas_read_analysis_config(configpath)
    print('Read config.')
    if awskeys['access_key'] == '' or awskeys['secret_key'] == '' or not 'user_group_name' in ncaasconfig.keys():
        print('NeuroCAAS credentials not found.')
        cred = CredentialsManager()
        app.exec_()
    from botocore.exceptions import ClientError
    try:
        awskeys = ncaas_read_aws_keys()
        s3_connect()
        wind = NCAASwrapper(folder = folder)
        sys.exit(app.exec_())
    except ClientError:
        print('Could not connect to NeuroCAAS, check credentials.')
        cred = CredentialsManager()
        app.exec_()
        awskeys = ncaas_read_aws_keys()
        s3_connect()
        wind = NCAASwrapper(folder = folder)
        sys.exit(app.exec_())
    



#  wfield - tools to analyse widefield data 
# Copyright (C) 2020 Joao Couto - jpcouto@gmail.com
#
#  This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .utils import *
from .io import *
from .registration import motion_correct
from .decomposition import svd_blockwise,approximate_svd,reconstruct
from .hemocorrection import hemodynamic_correction
from .allen import *
# other
from .imutils import *
from .viz import *
from .utils_svd import *
from .plots import *



from .utils import *
import yaml
import json
import threading

try:
    import boto3
except:
    print('boto3 not installed, installing with pip ')
    from subprocess import call
    call('pip install boto3',shell = True)
    import boto3    
defaultconfig = {
    'wfield-preprocess': dict(
        submit = dict(
            instance_type =  'r5.16xlarge',
            analysis_extension = '.zip',
            userfolder = 'ChurchlandGroup',
            decompress_results = True),  # this will decompress the U matrix when downloading
        config = dict(block_height = 180,
                      block_width = 160,
                      frame_rate = 60,
                      max_components = 5,
                      num_sims = 64,
                      num_channels = 2,
                      overlapping = True,
                      window_length = 1000)), # 7200    
    'cshl-wfield-locanmf': {
        'submit':dict(
            instance_type =  'p3.2xlarge',
            userfolder = 'data',
            params_filename = 'config.yaml',
            areanames_filename = 'labels.json',
            atlas_filename = 'atlas.npy',
            brainmask_filename = 'brainmask.npy',
            temporal_data_filename = 'SVTcorr.npy',
            spatial_data_filename = 'U_atlas.npy'),
        'config' :{
            "maxrank": 3,
            "loc_thresh": 80,
            "min_pixels": 100,
            "r2_thresh": 0.99,
	    "maxiter_hals":20,
	    "maxiter_lambda":300,
	    "lambda_step":1.35,
	    "lambda_init":0.000001}}}

def ncaas_set_aws_keys(access_key,secret_key,region='us-east-1'):
    fname = pjoin(os.path.expanduser('~'),'.aws','credentials')
    cred = '''[default]
aws_access_key_id = {access_key}
aws_secret_access_key = {secret_key}
'''
    dirname = os.path.dirname(fname)
    if not os.path.isdir(dirname):
        os.makedirs(dirname)    
    with open(fname,'w') as fd:
        fd.write(cred.format(access_key=access_key,secret_key = secret_key))
    fname = pjoin(os.path.expanduser('~'),'.aws','config')
    if not region is None:
        conf = '''[default]
region={region}'''
        with open(fname,'w') as fd:
            fd.write(conf.format(region=region))

def ncaas_read_aws_keys():
    awsfolder = pjoin(os.path.expanduser('~'),'.aws')
    awscredfile = pjoin(awsfolder,'credentials')
    awsconfig = pjoin(awsfolder,'credentials')
    access_key = ''
    secret_key = ''
    region = 'us-east-1'
    if os.path.isfile(awscredfile):
        with open(awscredfile,'r') as fd:
            for ln in fd:
                if 'aws_access_key_id' in ln:
                    ln = ln.split('=')
                    if len(ln)>1:
                        access_key = ln[-1].strip(' ').strip('\n')
                if 'aws_secret_access_key' in ln:
                    ln = ln.split('=')
                    if len(ln)>1:
                        secret_key = ln[-1].strip(' ').strip('\n')
    if os.path.isfile(awsconfig):
        with open(awsconfig,'r') as fd:
            for ln in fd:
                if 'region' in ln:
                    ln = ln.split('=')
                    region = ln[-1].strip(' ')
    return dict(access_key = access_key,
                secret_key = secret_key,
                region = region)    

def ncaas_read_analysis_config(configname):
    if not os.path.exists(os.path.dirname(configname)):
        os.makedirs(os.path.dirname(configname))
    '''
    if os.path.exists(configname):
        with open(configname,'r') as f:
            config = json.load(f)
            # This is to delete a config from a previous version if there.
        if not list(defaultconfig.keys())[0] in config.keys():
            os.remove(configname)
            dirname = os.path.dirname(configname)
            qfile = pjoin(dirname,'ncaas_transfer_q.json')
            if os.path.isfile(qfile):
                os.remove(qfile)
    '''
    if not os.path.exists(configname):
        with open(configname,'w') as f:
            print('Creating config from defaults [{0}]'.format(configname))
            json.dump(defaultconfig, f,
                      indent = 4,
                      sort_keys = True)
    with open(configname,'r') as f:
        config = json.load(f)
        for k in config.keys():
            if k in defaultconfig.keys(): # Use default
                for j in defaultconfig[k].keys():
                    if not j in config[k].keys():
                        config[k][j] = defaultconfig[k][j]
    return config

def get_tree_path(items,root = ''):
    ''' Get the paths from a QTreeView item'''
    paths = []
    for item in items:
        level = 0
        index = item
        paths.append([index.data()])
        while index.parent().isValid():
            index = index.parent()
            level += 1
            paths[-1].append(index.data())
        for i,p in enumerate(paths[-1]):
            if p is None :
                paths[-1][i] = ''
        paths[-1] = '/'.join(paths[-1][::-1])
    return paths

def make_tree(item, tree):
    if len(item) == 1:
        if not item[0] == '':
            tree[item[0]] = item[0]
    else:
        head, tail = item[0], item[1:]
        tree.setdefault(head, {})
        make_tree(
            tail,
            tree[head])


def to_log(msg,logfile = None):
    if logfile is None:
        logfile = open(pjoin(
                     os.path.expanduser('~'),
                     '.wfield','ncaas_gui_log.txt'),'a')
    nmsg = '['+datetime.today().strftime('%y-%m-%d %H:%M:%S')+'] - ' + msg + '\n'
    logfile.seek(os.SEEK_END)
    logfile.write(nmsg)
    return nmsg

def tail(filename, nlines=100):
    """
    This needs work (should not read the whole file).
    """
    with open(filename,'r') as f:
        lines = f.readlines()
    if len(lines) > 100:
        lines = lines[-100:]
    return lines


##########
### S3 ###
##########
awsregions = ['us-east-2',
              'us-east-1',
              'us-west-1',
              'us-west-2',
              'af-south-1',
              'ap-east-1',
              'ap-south-1',
              'ap-northeast-3',
              'ap-northeast-2',
              'ap-southeast-1',
              'ap-southeast-2',
              'ap-northeast-1',
              'ca-central-1',
              'cn-north-1',
              'cn-northwest-1',
              'eu-central-1',
              'eu-west-1',
              'eu-west-2',
              'eu-south-1',
              'eu-west-3',
              'eu-north-1',
              'me-south-1',
              'sa-east-1']


def s3_connect():
    aws = ncaas_read_aws_keys()
    session = boto3.session.Session(aws_access_key_id = aws['access_key'],
                                    aws_secret_access_key = aws['secret_key'])
    return session.resource('s3'), session.client('s3')


from PyQt5.QtWidgets import QApplication # Keep the gui as responsive as possible..

def s3_ls(s3client, s3, bucketnames, folder ):
    files = []
    for bucketname in bucketnames:
        #bucket = s3.Bucket(bucketname)
        #t = bucket.objects.filter(Delimiter = '/',Prefix='{0}/'.format(folder))
        f = []
        files.extend(s3_ls_r(f,
                             s3client,
                             bucketname,
                             folder+'/'))
        
    return files

# recursive ls
def s3_ls_r(files,s3client,bucket,prefix):
    try:
        objlist = s3client.list_objects_v2(Bucket=bucket,
                                           Delimiter = '/',
                                           Prefix=prefix)
    except:
        return files
    if 'Contents' in objlist.keys():
        for obj in objlist['Contents']:
            QApplication.processEvents()
            files.append(bucket + '/'+obj['Key'])
    if 'CommonPrefixes' in objlist.keys():
        for obj in objlist['CommonPrefixes']:
            QApplication.processEvents()
            files.append(bucket + '/'+obj['Prefix'])
            s3_ls_r(files,s3client,bucket,obj['Prefix'])
    return files


from boto3.s3.transfer import TransferConfig
multipart_config = TransferConfig(multipart_threshold=1024*25,
                                  max_concurrency=12,
                                  multipart_chunksize=1024*25,
                                  use_threads=True)

class Upload(threading.Thread):
    def __init__(self,filepath,destination,bucket,s3):
        super(Upload,self).__init__()
        self.s3 = s3
        statinfo = os.stat(filepath)
        self.fsize = statinfo.st_size
        self.count = 0
        self.isrunning = False
        self.bucket = bucket
        self.filepath = filepath
        self.destination = destination
    def run(self):
        self.isrunning = True
        print('Running upload on {0} to {2}/{1}'.format(self.filepath,
                                                        self.destination,
                                                        self.bucket))
        def update(chunk):
            self.count += chunk
            self.isrunning = True
        bucket =self.s3.Bucket(self.bucket)
        bucket.upload_file(self.filepath,
                           self.destination,
                           Callback = update,
                           Config = multipart_config)
        self.isrunning = False





#  wfield - tools to analyse widefield data - utils and general imports 
# Copyright (C) 2020 Joao Couto - jpcouto@gmail.com
#
#  This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os
import sys
import time
try:
    import cv2 # OpenCV needs to be imported before numpy for some seg faulted reason...
except:
    print('Some functionality might be broken: install opencv-python or opencv-python-headless')
import numpy as np
from tqdm import tqdm
from natsort import natsorted
from glob import glob
from os.path import join as pjoin
from datetime import datetime
from skimage.transform import warp
from multiprocessing import Pool, cpu_count
from functools import partial
from scipy.interpolate import interp1d
from scipy.sparse import load_npz, issparse,csr_matrix

print = partial(print, flush=True)


# create the wfield folder and place the references.
# this runs if installed from pip
wfield_dir = pjoin(os.path.expanduser('~'),'.wfield')

def _create_wfield_folder():
    if not os.path.isdir(wfield_dir):
        print('Created {0}'.format(wfield_dir))
        os.makedirs(wfield_dir)
    # try first from the shared folder 
    modulepath = pjoin(__file__.split('lib')[0],'share','wfield')
    refpath = pjoin(modulepath, 'references')
    if os.path.exists(refpath):
        reference_files = [pjoin(refpath,r) for r in os.listdir(refpath)]
        from shutil import copyfile
        for f in reference_files:
            if os.path.isfile(f):
                copyfile(f,f.replace(refpath,wfield_dir))
    else:
        import requests
        from shutil import copyfileobj
        webpath = 'https://raw.githubusercontent.com/jcouto/wfield/master/references/{0}'
        files = ['dorsal_cortex_ccf_labels.json',
                 'dorsal_cortex_landmarks.json',
                 'dorsal_cortex_outline.npy',
                 'dorsal_cortex_projection.npy',
                 'vis_ccf_labels.json',
                 'vis_outline.npy',
                 'vis_projection.npy']
        # download the files
        for f in files:
            print('    Downloading {0}'.format(f))
            if '.json' in f: # because of the encodings.
                with open(pjoin(wfield_dir,f),'w') as fid:
                    res = requests.get(webpath.format(f))
                    fid.write(res.text)
                    del res
            else:
                with open(pjoin(wfield_dir,f),'wb') as fid:
                    raw = requests.get(webpath.format(f),stream=True)
                    copyfileobj(raw.raw, fid)
                    del raw

def estimate_similarity_transform(ref,points):
    '''
    
    ref = np.vstack([landmarks_im['x'],landmarks_im['y']]).T
    match = point_stream.data    
    cor = np.vstack([match['x'],match['y']]).T
    
    M = estimate_similarity_transform(ref, cor)
    '''
    from skimage.transform import SimilarityTransform
    M = SimilarityTransform()
    M.estimate(ref,points)
    return M 

def im_adapt_hist(im,clip_limit = .1, grid_size=(8,8)):
    ''' Adaptative histogram of image

        eqim = im_adapt_hist(im,.1)
    '''
    clahe = cv2.createCLAHE(clipLimit=clip_limit, tileGridSize=grid_size)
    return clahe.apply(im.squeeze())

def im_apply_transform(im,M,dims = None):
    '''
    Applies an affine transform M to an image.
    nim = im_apply_transform(im,M)

    '''
    if issparse(im):
        # then reshape before
        if dims is None:
            raise ValueError('Provide dims when warping sparse matrices.')
        shape = im.shape
        tmp  = np.asarray(im.todense()).reshape(dims)
        tmp = warp(tmp,M,
                   order = 1,
                   mode='constant',
                   cval = 0,
                   clip = True,
                   preserve_range = True)
        return csr_matrix(tmp.reshape(shape))
    else:    
        return warp(im,M,
                    order = 1,
                    mode='constant',
                    cval = 0,
                    clip=True,
                    preserve_range=True)

def lowpass(X, w = 7.5, fs = 30.):
    from scipy.signal import butter, filtfilt
    b, a = butter(2,w/(fs/2.), btype='lowpass');
    return filtfilt(b, a, X, padlen = 50)

def highpass(X, w = 3., fs = 30.):
    from scipy.signal import butter, filtfilt
    b, a = butter(2,w/(fs/2.), btype='highpass');
    return filtfilt(b, a, X, padlen = 50)


def analog_ttl_to_onsets(dat,time=None, mfilt=3):
    if time is None:
        time = np.arange(len(dat))
    if mfilt:
        from scipy.signal import medfilt
        dat = medfilt(dat,mfilt)
    tt = np.diff(dat.astype(np.float32))
    onsets = np.where(tt-np.max(np.abs(tt))/2 > 0)[0]+1
    offsets = np.where(tt+np.max(np.abs(tt))/2 < 0)[0]+1
    return time[onsets],time[offsets]

def chunk_indices(nframes, chunksize = 512, min_chunk_size = 16):
    '''
    Gets chunk indices for iterating over an array in evenly sized chunks
    '''
    chunks = np.arange(0,nframes,chunksize,dtype = int)
    if (nframes - chunks[-1]) < min_chunk_size:
        chunks[-1] = nframes
    if not chunks[-1] == nframes:
        chunks = np.hstack([chunks,nframes])
    return [[chunks[i],chunks[i+1]] for i in range(len(chunks)-1)]


def make_overlapping_blocks(dims, blocksize = 128, overlap = 16):
    '''
    Creates overlapping block indices to span an image
    '''
    
    w,h=dims
    blocks = []
    for i,a in enumerate(range(0,w,blocksize-overlap)):
        for j,b in enumerate(range(0,h,blocksize-overlap)):
            blocks.append([(a,np.clip(a+blocksize,0,w)),(b,np.clip(b+blocksize,0,h))])
    return blocks

def reconstruct(u,svt,dims = None):
    if issparse(u):
        if dims is None:
            raise ValueError('Supply dims = [H,W] when using sparse arrays')
    else:
        if dims is None:
            dims = u.shape[:2]
    return u.dot(svt).reshape((*dims,-1)).transpose(-1,0,1).squeeze()


class SVDStack(object):
    def __init__(self, U, SVT, dims = None, warped = None, dtype = 'float32'):
        self.U = U.astype('float32')
        self.SVT = SVT.astype('float32')
        self.issparse = False
        if issparse(U):
            self.issparse = True
            if dims is None:
                raise ValueError('Supply dims = [H,W] when using sparse arrays')
            self.Uflat = self.U
        else:
            if dims is None:
                dims = U.shape[:2]
            self.Uflat = self.U.reshape(-1,self.U.shape[-1])
        self.U_warped = warped
        self.warped = False
        self.M = None
        self.shape = [SVT.shape[1],*dims]
        self.dtype = dtype
        self.originalU = None
        
    def set_warped(self,value,M = None):
        ''' Apply affine transform to the spatial components '''
        if not M is None:
            self.M = M
        if self.originalU is None:
            self.originalU = self.U.copy()
        if not value:
            self.U = self.originalU
            self.warped = False
        else:
            if self.U_warped is None:
                if not self.M is None:
                    if not self.issparse:
                        if self.originalU is None:
                            self.originalU = self.U.copy()
                        self.U_warped = self.originalU.copy()
                        self.U_warped[:,0,:] = 1e-10
                        self.U_warped[0,:,:] = 1e-10
                        self.U_warped[-1,:,:] = 1e-10
                        self.U_warped[:,-1,:] = 1e-10
                        self.U_warped = np.stack(runpar(im_apply_transform,
                                                        self.U_warped.transpose([2,0,1]),
                                                        M = self.M)).transpose([1,2,0]).astype(np.float32)
            if not self.U_warped is None:
                self.U = self.U_warped
                self.warped = True
        self.Uflat = self.U.reshape(-1,self.U.shape[-1])
        return
    def __len__(self):
        return self.SVT.shape[1]
    def __getitem__(self,*args):
        ndims  = len(args)
        if type(args[0]) is slice:
            idxz = range(*args[0].indices(self.shape[0]))
        else:
            idxz = args[0]        
        return reconstruct(self.U,self.SVT[:,idxz],dims = self.shape[1:])
    def get_timecourse(self,xy):
        # index are in xy, like what np.where(mask) returns
        x = np.array(np.clip(xy[0],0,self.shape[1]),dtype=int)
        y = np.array(np.clip(xy[1],0,self.shape[2]),dtype=int)
        idx = np.ravel_multi_index((x,y),self.shape[1:])
        t = self.Uflat[idx,:].dot(self.SVT)
        return t

def get_trial_baseline(idx,frames_average,onsets):
    if len(frames_average.shape) <= 3:
        return frames_average
    else:
        if onsets is None:
            print(' Trial onsets not defined, using the first trial')
            return frames_average[0]
        return frames_average[np.where(onsets<=idx)[0][-1]]

def point_find_ccf_region(point,ccf_regions,sides = ['left','right'],approx_value=-0.01):
    '''
    Find the area where a point is contained from ccf_regions contours.
    Example:
    point = [6.5,3.0]
    region,side,index = point_find_ccf_region(point,refregions)
    
    '''
    region = None
    idx = None
    ir = None
    for ir, r in ccf_regions.iterrows():
        for s in sides:
            c = np.vstack([r[s + '_x'], r[s + '_y']]).astype('float32').T
            c = c.reshape([c.shape[0],1,c.shape[1]])
            d = cv2.pointPolygonTest(c,tuple(point),True)
            if d >= approx_value:
                region = r
                idx = ir
                side = s
        if not region is None:
            break
    return region,side,idx

def contour_to_im(x,y,dims,extent=None,n_up_samples = 1000):
    '''
    im = contour_to_im(x,y,dims,extent=None,n_up_samples = 1000)
    
    Imprint a contour on an image.
    
    see also: contour_to_mask
'''
    if extent is None:
        extent = [0,dims[0],0,dims[1]]
    
    cont = np.vstack([y,x]).T
    x = np.linspace(extent[0], extent[1], dims[0]+1)
    y = np.linspace(extent[2], extent[3], dims[1]+1)
    
    C = cont.copy()
    C = np.vstack([C,C[0,:]])
    if n_up_samples > cont.shape[0]:
        C = np.zeros((n_up_samples,2))
        C[:,0] = interp1d(np.linspace(0,1,cont.shape[0]),
                          np.clip(cont[:,0],0,dims[0]))(
                              np.linspace(0,1,n_up_samples))
        C[:,1] = interp1d(np.linspace(0,1,cont.shape[0]),
                          np.clip(cont[:,1],0,dims[1]))(
                              np.linspace(0,1,n_up_samples))
    C[:,0] = np.clip(C[:,0],0,dims[0])
    C[:,1] = np.clip(C[:,1],0,dims[1])
    H, xedges, yedges = np.histogram2d(C[:,0], C[:,1], bins=(np.sort(x), np.sort(y)))
    H = H>0
    return H.astype(bool)

def contour_to_mask(x,y,dims,extent = None,n_up_samples = 2000):
    '''
        H = contour_to_mask(x,y,dims,extent = None,n_up_samples = 2000)
    
        Create a mask from a contour
        
    '''

    H = contour_to_im(x=x, y=y, 
                      dims = dims,
                      extent = extent,
                      n_up_samples = n_up_samples)    
    # fix border cases
    #if np.sum(H[0,:]):
    #    H[0,:] = np.uint8(1)
    #if np.sum(H[-1,:]):
    #    H[-1,:] = np.uint8(1)
    #if np.sum(H[:,0]):
    #    H[:,0] = np.uint8(1)
    #if np.sum(H[:,-1]):
    #    H[:,-1] = np.uint8(1)
    from scipy.ndimage import morphology
    H = morphology.binary_dilation(H)
    H = morphology.binary_fill_holes(H)
    H = morphology.binary_erosion(H)
    H[0,:] = 0
    H[-1,:] = 0
    H[:,0] = 0
    H[:,-1] = 0
    return H.astype(bool)


def parinit():
    import os
    os.environ['MKL_NUM_THREADS'] = "1"
    os.environ['OMP_NUM_THREADS'] = "1"

def runpar(f,X,nprocesses = None,**kwargs):
    ''' 
    res = runpar(function,          # function to execute
                 data,              # data to be passed to the function
                 nprocesses = None, # defaults to the number of cores on the machine
                 **kwargs)          # additional arguments passed to the function (dictionary)

    '''
    if nprocesses is None:
        nprocesses = cpu_count()
    with Pool(initializer = parinit, processes=nprocesses) as pool:
        res = pool.map(partial(f,**kwargs),X)
    pool.join()
    return res


def zipdir(path, outputpath):
    import zipfile
    with zipfile.ZipFile(outputpath, 'w', 
                         allowZip64 = True) as zipf:   
        for root, dirs, files in os.walk(path):
            for file in tqdm(files,desc = 'Compressing'):
                zipf.write(os.path.join(root, file),file)

def zipfiles(files, outputpath):
    import zipfile
    with zipfile.ZipFile(outputpath, 'w', 
                         allowZip64 = True) as zipf:   
        for f in tqdm(files,desc = 'Compressing'):
            zipf.write(f,os.path.basename(f))



#!/usr/bin/env python
# Install script for widefield tools

#  wfield - tools to analyse widefield data
# Copyright (C) 2020 Joao Couto - jpcouto@gmail.com
#
#  This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from setuptools import setup
from setuptools.command.install import install
import os
from os.path import join as pjoin
from distutils.cmd import Command

description = '''Utilities to look at widefield data and align with the allen reference map.'''

with open("README.md", "r") as fh:
    longdescription = fh.read()

requirements = []
with open("requirements.txt","r") as f:
    requirements = f.read().splitlines()
    
setup(
    name = 'wfield',
    version = '0.3.4',
    author = 'Joao Couto',
    author_email = 'jpcouto@gmail.com',
    description = (description),
    long_description = longdescription,
    long_description_content_type='text/markdown',
    license = 'GPL',
    install_requires = requirements,
    url = "https://github.com/jcouto/wfield",
    packages = ['wfield'],
    entry_points = {
        'console_scripts': [
            'wfield = wfield.cli:main',
            'wfield-ncaas = wfield.ncaas_gui:main',
        ],
        
    })



#  wfield - tools to analyse widefield data - motion correction 
# Copyright (C) 2020 Joao Couto - jpcouto@gmail.com
#
#  This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from  .utils import *
from skimage.transform import AffineTransform
cv2.setNumThreads(1)

def findTransformECC(template,dst,M,warp_mode,criteria,inputMask,gaussFiltSize):
    return cv2.findTransformECC(template,dst,
                                M, warp_mode,
                                criteria,
                                inputMask=inputMask,
                                gaussFiltSize=gaussFiltSize)

cv2ver = cv2.__version__.split('.')
if (int(cv2ver[0]) == 3) and (int(cv2ver[1]) <= 4):
    if int(cv2ver[2]) <= 5:
        def findTransformECC(template,
                             dst,
                             M,
                             warp_mode,
                             criteria,
                             inputMask,
                             gaussFiltSize):
            return cv2.findTransformECC(template,dst,
                                        M, warp_mode,
                                        criteria,
                                        inputMask=inputMask)
elif (int(cv2ver[0]) == 4) and (int(cv2ver[1]) <= 1):
    # gaussFiltSize is a mandatory input on opencv 4.4 but not 4.1
    def findTransformECC(template,
                         dst,
                         M,
                         warp_mode,
                         criteria,
                         inputMask,
                         gaussFiltSize):
        return cv2.findTransformECC(template,dst,
                                    M, warp_mode,
                                    criteria,
                                    inputMask=inputMask)
def registration_ecc(frame,template,
                     niter = 100,
                     eps0 = 1e-3,
                     warp_mode = cv2.MOTION_EUCLIDEAN,
                     prepare = True,
                     gaussian_filter = 1,
                     hann = None,
                     **kwargs):
    h,w = template.shape
    if hann is None:
        hann = cv2.createHanningWindow((w,h),cv2.CV_32FC1)
        hann = (hann*255).astype('uint8')
    dst = frame.astype('float32')
    M = np.eye(2, 3, dtype=np.float32)
    criteria = (cv2.TERM_CRITERIA_EPS | cv2.TERM_CRITERIA_COUNT,
                niter,  eps0)
    (res, M) = findTransformECC(template,dst,
                                M, warp_mode,
                                criteria,
                                inputMask=hann, gaussFiltSize=gaussian_filter)
    dst = cv2.warpAffine(frame, M, (w,h),
                         flags=cv2.INTER_LINEAR + cv2.WARP_INVERSE_MAP);
    return M, np.clip(dst,0,(2**16-1)).astype('uint16')


def _xy_rot_from_affine(affines):
    '''
    helper function to parse affine parameters from ECC
    '''
    xy = []
    rot = []
    for r in affines:
        M = np.vstack([r, np.array([0,0,1])])
        M = AffineTransform(M)
        xy.append(M.translation)
        rot.append(M.rotation)
    rot = np.rad2deg(np.array(rot))
    xy = np.array(xy)
    return xy,rot

def registration_upsample(frame,template):
    h,w = frame.shape
    dst = frame.astype('float32')
    (xs, ys), sf = cv2.phaseCorrelate(dst,template.astype('float32'))    
    M = np.float32([[1,0,xs],[0,1,ys]])
    dst = cv2.warpAffine(dst,M,(w, h))
    return (xs,ys),(np.clip(dst,0,(2**16-1))).astype('uint16')

def _register_multichannel_stack(frames,templates,mode='2d',
                                 niter = 100,
                                 eps0 = 1e-3,
                                 warp_mode = cv2.MOTION_EUCLIDEAN): # mode 2d

    nframes, nchannels, h, w = frames.shape     
    if mode == 'ecc':
        hann = cv2.createHanningWindow((w,h),cv2.CV_32FC1)
        hann = (hann*255).astype('uint8')

    ys = np.zeros((nframes,nchannels),dtype=np.float32)
    xs = np.zeros((nframes,nchannels),dtype=np.float32)
    rot = np.zeros((nframes,nchannels),dtype=np.float32)
    stack = np.zeros_like(frames,dtype = 'uint16')
    for ichan in range(nchannels):
        chunk = frames[:,ichan].squeeze()
        if mode == '2d':
            res = runpar(registration_upsample, chunk,
                         template = templates[ichan])
            ys[:,ichan] = np.array([r[0][1] for r in res],dtype='float32')
            xs[:,ichan] = np.array([r[0][0] for r in res],dtype='float32')

        elif mode == 'ecc':
            res = runpar(registration_ecc, chunk,
                         template = templates[ichan],
                         hann = hann,
                         niter = niter,
                         eps0 = eps0,
                         warp_mode = warp_mode)
            xy,rots = _xy_rot_from_affine([r[0] for r in res])
            ys[:,ichan] = xy[:,1]
            xs[:,ichan] = xy[:,0]
            rot[:,ichan] = rots
        stack[:,ichan,:,:] = np.stack([r[1] for r in res])
    return (xs,ys,rot), stack

def motion_correct(dat, out = None,
                   refs = None,
                   chunksize=512,
                   nreference = 60,
                   mode = 'ecc',
                   apply_shifts=True):
    '''
    Motion correction by translation.
    This estimate x and y shifts using phase correlation. 
    
    The reference image is the average of the chunk in the center.

    Inputs:
        dat (array)           : (NFRAMES, NCHANNEL, H, W) is overwritten if apply_shifts is True
        out (array)           : same size as dat or None to overwrite dat
        refs (array)          : reference frames (NCHANNEL, H, W) or None to compute from nreference frames
        chunksize (int)       : size of the chunks (needs to be small enough to fit in memory - default 512)
        nreference            : number of frames to take as reference (default 60)
        apply_shifts          : overwrite the data with the motion corrected (default True)
        mode                  : ecc (default) is rigid body; 2d is only translation in x and y using dft
    Returns:
        (yshifts, xshifts)    : shitfs in y and x ((NFRAMES, NCHANNELS),(NFRAMES, NCHANNELS))
        rot_shifts            : rotational shifts if in ecc mode 
    '''
    nframes,nchan,h,w = dat.shape
    if out is None:
        out = dat
    chunks = chunk_indices(nframes,chunksize)
    xshifts = []
    yshifts = []
    rshifts = []
    # reference is from the start of the file (nreference frames to nreference*2)
    # (chunksize frames and for each channel independently)
    if refs is None:
        nreference = int(nreference)
        chunk = np.array(dat[nreference:nreference*2])
        refs = chunk[0].astype('float32')
        # align to the ref of each channel and use the mean
        _,refs = _register_multichannel_stack(chunk,refs,mode=mode)
        refs = np.mean(refs,axis=0).astype('float32')
    for c in tqdm(chunks,desc='Motion correction'):
        # this is the reg bit
        localchunk = np.array(dat[c[0]:c[-1]])
        (xs,ys,rot),corrected = _register_multichannel_stack(
            localchunk,
            refs,
            mode=mode)
        if apply_shifts:
            out[c[0]:c[-1]] = corrected[:]
            if hasattr(out,'flush'):
                out.flush() # write to disk
        yshifts.append(ys)
        xshifts.append(xs)
        rshifts.append(rot)
    return (np.vstack(yshifts),np.vstack(xshifts)),np.vstack(rshifts)





#  wfield - tools to analyse widefield data - visualization 
# Copyright (C) 2020 Joao Couto - jpcouto@gmail.com
#
#  This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pylab as plt
from .utils import *

def two_chan_to_rgb(dat,norm=True):
    img = np.stack([*dat,
                          np.zeros_like(dat[0])],
                   axis = 0)
    if norm:
        for i in range(len(dat)):
            img[i] /= img[i].max()
        img *= 255
        img = img.astype('uint8')
    img = img.transpose([1,2,0])
    return img


# Settings for colors and saving plots
colors = ['#000000',
          '#d62728',
          '#1f77b4',
          '#ff7f0e',
          '#2ca02c',
          '#9467bd',
          '#8c564b',
          '#e377c2',
          '#7f7f7f',
          '#bcbd22',
          '#17becf']

mplstyledict = {'pdf.fonttype' : 42,
                'ps.fonttype' : 42,
                'svg.fonttype':'none',
                'axes.titlesize' : 14,
                'axes.labelsize' : 12,
                'axes.facecolor': 'none',
                'axes.linewidth' : 0.8,
                'axes.spines.right': False,
                'axes.spines.top': False,
                'axes.titleweight': 'bold',
                'axes.prop_cycle': plt.cycler('color', colors),
                'ytick.major.size': 8,
                'xtick.major.size': 8,
                'xtick.labelsize' : 8,
                'ytick.labelsize' : 8,
                'xtick.major.width': 1,
                'ytick.major.width': 1,
                'figure.edgecolor': 'none',
                'figure.facecolor': 'none',
                'figure.frameon': False,
                'font.family': ['sans-serif'],
                'font.size'  : 12,
                'font.sans-serif': ['Arial',
                                    'DejaVu Sans',
                                    'Bitstream Vera Sans',
                                    'Computer Modern Sans Serif',
                                    'Lucida Grande',
                                    'Verdana',
                                    'Geneva',
                                    'Lucid',
                                    'Avant Garde',
                                    'sans-serif'],             
                'lines.linewidth' : 1.5,
                'lines.markersize' : 4,
                'image.interpolation': 'none',
                'image.resample': False,
                'savefig.dpi': 500
}
plt.matplotlib.style.use(mplstyledict)

################################################################
#######################  MATPLOTLIB WRAPPERS ###################
################################################################

def imshow_noborder(img,fig = None,figsize = [7,7],**kwargs):
    if fig is None:
        fig = plt.figure(figsize = figsize)
        ax = fig.add_axes([0,0,1,1])
    plt.gca().axis('off')            
    plt.imshow(img,**kwargs)
    return fig

###############################################################
#######################  NOTEBOOK WRAPPERS  ###################
###############################################################

def _handle_sparse(im,shape):
    if issparse(im):
        if shape is None:
            raise ValueError('Supply shape = [H,W] when using sparse arrays')
        im = np.asarray(im.todense()).reshape(shape)
    return im

def nb_play_movie(data,interval=30,shape = None,**kwargs):
    ''' 
    Play a movie from the notebook
    '''
    from ipywidgets import Play,jslink,HBox,IntSlider
    from IPython.display import display

    i = _handle_sparse(data[0],shape = shape)
    im = plt.imshow(i.squeeze(),**kwargs)
    slider = IntSlider(0,min = 0,max = data.shape[0]-1,step = 1,description='Frame')
    play = Play(interval=interval,
                value=0,
                min=0,
                max=data.shape[0]-1,
                step=1,
                description="Press play",
                disabled=False)
    jslink((play, 'value'), (slider, 'value'))
    display(HBox([play, slider]))
    def updateImage(change):
        i = _handle_sparse(data[change['new']],shape=shape)
        im.set_data(i.squeeze())
        plt.gcf().canvas.draw()
        plt.gcf().canvas.flush_events()
    slider.observe(updateImage, names='value')
    return dict(fig = plt.gcf(),
                ax=plt.gca(),
                im= im,
                update = updateImage)

def nb_save_movie(data,filename,interval = 100,dpi = 90,shape=None,**kwargs):
    '''
    Replace nb_play_movie with this to save to a file.

    Example:

    nb_save_movie(tmp[:,:,::-1],
                  filename = '~/Desktop/example.avi',
                  clim = [.06,.2],
                  extent = extent,
                  cmap = 'hot', 
                  alpha = 0.5);
    '''
    from tqdm import tqdm
    from matplotlib.animation import FuncAnimation
    def animate(frame):
        global pbar
        pbar.update(1)
        i = _handle_sparse(data[frame],shape = shape)
        im.set_data(i.squeeze())        
        return im,
    fig = plt.gcf()
    i = _handle_sparse(data[0],shape = shape)
    im = plt.imshow(i.squeeze(),**kwargs)
    animation = FuncAnimation(
        fig,
        animate,
        np.arange(data.shape[0]),
        fargs=[],
        interval=interval)
    global pbar
    pbar = tqdm(desc = 'Saving movie ',total=data.shape[0])
    animation.save(filename, dpi=dpi)
    pbar.close()
    plt.show()
    print('Saved to {0}'.format(filename))



################################################################
##################### PYQTGRAPH WRAPPERS #######################
################################################################

def qtgraph_show_svd(stack):
    from .widgets import SVDViewer
    return SVDViewer(stack)

################################################################
##################### HOLOVIEWS WRAPPERS #######################
################################################################

def hv_imshow_stack(X,cmap = 'gray',scale=.8,title='dataset',timelabel = 'frame'):
    import holoviews as hv
    hv.extension('bokeh')
    d = X.shape
    if len(d) == 4:
        ds = hv.Dataset((np.arange(d[3]), np.arange(d[2]), np.arange(d[1]),np.arange(d[0]),
                         X),
                        ['h','w', 'ch', timelabel], title)
    elif len(d) == 3:
        ds = hv.Dataset((np.arange(d[2]), np.arange(d[1]),np.arange(d[0]),
                         X),
                        ['h','w', timelabel], title)
    else:
        print('imshow_stack only works for 3d and 4d stacks.')
        return None
    im = ds.to(hv.Image, ['h', 'w'],dynamic=True)
    im.opts(cmap=cmap,width=int(d[-1]*scale),height=int(d[-2]*scale))
    return im

################################################################
#######################  NAPARI WRAPPERS #######################
################################################################


def napari_show(dat,contrast_limits = None):
    import napari
    if contrast_limits is None:
        contrast_limits = [dat[0].min(),dat[0].max()]
    with napari.gui_qt():
        try:
            napari.view_image(dat,
                              contrast_limits=contrast_limits,
                              is_pyramid=False)
        except: # napari 0.3.5 .... why??
            napari.view_image(dat,
                              contrast_limits=contrast_limits)
            



#  wfield - tools to analyse widefield data - hemodynamics correction 
# Copyright (C) 2020 Joao Couto - jpcouto@gmail.com
#
#  This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .utils import *
# Pass temporal components in channels
# Uses a sparse U


def _hemodynamic_find_coeffs(U,SVTa,SVTb):
    a = np.dot(U,SVTa)
    b = np.dot(U,SVTb)
    eps = 1.e-10 # so we never divide by zero
    return np.nansum(a*b,axis = 1)/(np.nansum(b*b,axis = 1)+eps)

def hemodynamic_correction(U, SVT_470,SVT_405,
                           fs=30.,
                           freq_lowpass = 14.,
                           freq_highpass = 0.1,
                           nchunks = 1024,
                           run_parallel = True):
    # split channels and subtract the mean to each
    SVTa = SVT_470#[:,0::2]
    SVTb = SVT_405#[:,1::2]

    # reshape U
    dims = U.shape
    U = U.reshape([-1,dims[-1]])

    # Single channel sampling rate
    fs = fs
    # Highpass filter
    if not freq_highpass is None:
        SVTa = highpass(SVTa,w = freq_highpass, fs = fs)
        SVTb = highpass(SVTb,w = freq_highpass, fs = fs)
    if not freq_lowpass is None:
        if freq_lowpass < fs/2:
            SVTb = lowpass(SVTb,freq_lowpass, fs = fs)
        else:
            print('Skipping lowpass on the violet channel.')
    # subtract the mean
    SVTa = (SVTa.T - np.nanmean(SVTa,axis=1)).T.astype('float32')
    SVTb = (SVTb.T - np.nanmean(SVTb,axis=1)).T.astype('float32')

    npix = U.shape[0]
    idx = np.array_split(np.arange(0,npix),nchunks)
    # find the coefficients
    if run_parallel:     # run in parallel
        rcoeffs = runpar(_hemodynamic_find_coeffs,
                         [U[ind,:] for ind in idx],
                         SVTa=SVTa,
                         SVTb=SVTb)
        rcoeffs = np.hstack(rcoeffs).astype('float32')
    else:                # run in series
        rcoeffs = np.zeros((npix))
        for i,ind in tqdm(enumerate(idx)):
            # rcoeffs[ind] = _hemodynamic_find_coeffs(U[ind,:],SVTa,SVTb)
            a = np.dot(U[ind,:],SVTa)
            b = np.dot(U[ind,:],SVTb)
            rcoeffs[ind] = np.sum(a*b,axis = 1)/np.sum(b*b,axis = 1)
    # drop nan
    rcoeffs[np.isnan(rcoeffs)] = 1.e-10
    # find the transformation
    T = np.dot(np.linalg.pinv(U),(U.T*rcoeffs).T)
    # apply correction
    SVTcorr = SVTa - np.dot(T,SVTb)
    # return a zero mean SVT
    SVTcorr = (SVTcorr.T - np.nanmean(SVTcorr,axis=1)).T.astype('float32')
    # put U dims back in case its used sequentially
    U = U.reshape(dims)
    
    return SVTcorr.astype('float32'), rcoeffs.astype('float32').reshape(dims[:2]), T.astype('float32')



#  wfield - tools to analyse widefield data - decomposition 
# Copyright (C) 2020 Joao Couto - jpcouto@gmail.com
#
#  This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .utils import *
from .io import load_binary_block
from numpy.linalg import svd

def approximate_svd(dat, frames_average,
                    onsets = None,
                    k=200, 
                    nframes_per_bin = 30,
                    nbinned_frames = 5000,
                    nframes_per_chunk = 500):
    '''
    Approximate single value decomposition by estimating U from the average movie and using it to compute S.VT.
    This is similar to what described in Steinmetz et al. 2017

    Joao Couto - March 2020
    
    This computes the mean centered SVD of the dataset, it does not compute F-F0/F0 a.k.a. df/f.
    Compute it after using the SVD components.

    Inputs:
        dat (array)             : (NFRAMES, NCHANNEL, H, W) 
        k (int)                 : number of components to estimate (200)
        nframes_per_bin (int)   : number of frames to estimate the initial U components
        nbinned_frames (int)    : maximum number frames to estimate tje initial U components
        nframes_per_chunk (int) : window size to load to memory each time.   
    Returns:
        U   (array)             : 
        SVT (array)             : 
    '''
    from sklearn.preprocessing import normalize

    if hasattr(dat,'filename'):
        dat_path = dat.filename
    else:
        dat_path = None
    dims = dat.shape[1:]

    # the number of bins needs to be larger than k because of the number of components.
    if nbinned_frames < k:
        nframes_per_bin = np.clip(int(np.floor(len(dat)/(k))),1,nframes_per_bin)

    nbinned_frames = np.min([nbinned_frames,
                             int(np.floor(len(dat)/nframes_per_bin))])
    
    idx = np.arange(0,nbinned_frames*nframes_per_bin,nframes_per_bin,
                    dtype='int')
    if not idx[-1] == len(dat):
        idx = np.hstack([idx,len(dat)-1])
    binned = np.zeros([len(idx)-1,*dat.shape[1:]],dtype = 'float32')
    for i in tqdm(range(len(idx)-1),desc='Binning raw data'):
        if dat_path is None:
            blk = dat[idx[i]:idx[i+1]] # work when data are loaded to memory
        else:
            blk = load_binary_block((dat_path,idx[i],nframes_per_bin),
                                    shape=dims)
        avg = get_trial_baseline(idx[i],frames_average,onsets)
        binned[i] = np.mean((blk-avg + np.float32(1e-5))
                            / (avg + np.float32(1e-5)), axis=0)
    binned = binned.reshape((-1,np.multiply(*dims[-2:])))

    # Get U from the single value decomposition 
    cov = np.dot(binned,binned.T)/binned.shape[1]
    cov = cov.astype('float32')

    u,s,v = svd(cov)
    U = normalize(np.dot(u[:,:k].T, binned),norm='l2',axis=1)
    k = U.shape[0]     # in case the k was smaller (low var)
    # if trials are defined, then use them to chunck data so that the baseline is correct
    if onsets is None:
        idx = np.arange(0,len(dat),nframes_per_chunk,dtype='int')
    else:
        idx = onsets
    if not idx[-1] == len(dat):
        idx = np.hstack([idx,len(dat)-1])
    V = np.zeros((k,*dat.shape[:2]),dtype='float32')
    # Compute SVT
    for i in tqdm(range(len(idx)-1),desc='Computing SVT from the raw data'):
        if dat_path is None:
            blk = dat[idx[i]:idx[i+1]] # work when data are loaded to memory
        else:
            blk = load_binary_block((dat_path,idx[i],idx[i+1]-idx[i]),
                                shape=dims).astype('float32')
        avg = get_trial_baseline(idx[i],frames_average,onsets).astype('float32')
        blk = (blk-avg+np.float32(1e-5))/(avg+np.float32(1e-5))
        V[:,idx[i]:idx[i+1],:] = np.dot(
            U, blk.reshape([-1,np.multiply(*dims[1:])]).T).reshape((k,-1,dat.shape[1]))  


    SVT = V.reshape((k,-1))
    U = U.T.reshape([*dims[-2:],-1])
    return U,SVT


def svd_blockwise(dat,frames_average,
                  k = 200, block_k = 20,
                  blocksize=120, overlap=8, 
                  random_state=42):
    '''
    Computes the blockwise single value decomposition for a matrix that does not fit in memory.

    U,SVT,S,(block_U,block_SVT,blocks) = svd_blockwise(dat,
                                                   frames_average,
                                                   k = 200, 
                                                   block_k = 20,
                                                   blocksize=120, 
                                                   overlap=8)
    dat is a [nframes X nchannels X width X height] array
    frames_average is a [nchannels X width X height] array; the average to be subtracted before computing the SVD
    k is the number of components to be extracted (randomized SVD)

    The blockwise implementation works by first running the SVD on overlapping chunks of the movie. Secondly,  SVD is ran on the extracted temporal components and the spatial components are scaled to match the actual frame size. 
The chunks have all samples in time but only a fraction of pixels.

    This is adapted from matlab code by Simon Musall.
    A similar approach is described in Stringer et al. Science 2019.

    Joao Couto - March 2020
    '''
    from sklearn.utils.extmath import randomized_svd
    from sklearn.preprocessing import normalize

    nframes,nchannels,w,h = dat.shape
    n = nframes*nchannels
    # Create the chunks where the SVD is ran initially, 
    #these have all samples in time but only a few in space 
    #chunks contain pixels that are nearby in space 
    blocks = make_overlapping_blocks((w,h),blocksize=blocksize,overlap=overlap)
    nblocks = len(blocks)
    # M = U.S.VT
    # U are the spatial components in this case
    block_U = np.zeros((nblocks,blocksize,blocksize,block_k),dtype=np.float32)
    block_U[:] = np.nan
    # V are the temporal components
    block_SVT = np.zeros((nblocks,block_k,n),dtype=np.float32)
    block_U[:] = np.nan
    # randomized svd is ran on each chunk 
    for iblock,(i,j) in tqdm(enumerate(blocks), total= len(blocks),
                             desc= 'Computing SVD on data chunks:'):
        # subtract the average (this should be made the baseline instead)
        arr = np.array(dat[:,:,i[0]:i[1],j[0]:j[1]],dtype='float32')
        arr -= frames_average[:,i[0]:i[1],j[0]:j[1]]
        arr /= frames_average[:,i[0]:i[1],j[0]:j[1]]
        bw,bh = arr.shape[-2:]
        arr = arr.reshape([-1,np.multiply(*arr.shape[-2:])])
        u, s, vt = randomized_svd(arr.T,
                                  n_components=block_k,
                                  n_iter=5,
                                  power_iteration_normalizer ='LQ',
                                  random_state=random_state)
        block_U[iblock,:bw,:bh,:] = u.reshape([bw,bh,-1])
        block_SVT[iblock] = np.dot(np.diag(s),vt)

    U,SVT,S = _complete_svd_from_blocks(block_U,block_SVT,blocks,k,(w,h))    
    return U,SVT,S,(block_U,block_SVT,blocks)      


def _complete_svd_from_blocks(block_U,block_SVT,blocks,k,dims,
                              n_iter=15,random_state=42):
    # Compute the svd of the temporal components from all blocks
    from sklearn.utils.extmath import randomized_svd
    u, s, vt = randomized_svd(
        block_SVT.reshape([np.multiply(*block_SVT.shape[:2]),-1]),
        n_components=k,
        n_iter=n_iter,
        power_iteration_normalizer ='QR',
        random_state=random_state)
    S = s;
    SVT = np.dot(np.diag(S),vt)
    # Map the blockwise spatial components compontents to the second SVD 
    U = np.dot(assemble_blockwise_spatial(block_U,blocks,dims),u)
    return U,SVT,S


def assemble_blockwise_spatial(block_U,blocks,dims):
    w,h = dims
    U = np.zeros([block_U.shape[0],block_U.shape[-1],w,h],dtype = 'float32')
    weights = np.zeros((w,h),dtype='float32')
    for iblock,(i,j) in enumerate(blocks):
        lw,lh = (i[1]-i[0],j[1]-j[0])
        U[iblock,:,i[0]:i[1],j[0]:j[1]] = block_U[iblock,:lw,:lh,:].transpose(-1,0,1)
        weights[i[0]:i[1],j[0]:j[1]] += 1
    U = (U/weights).reshape((np.multiply(*U.shape[:2]),-1))
    return U.T


#############################
############ PMD ############
#############################

# This was written by Ian Kinsella for the NeuroCAAS implementation
# of the penalized matrix factorization of widefield data with 2 channels.
# The interface was simplified to be used locally
# It may require alot of memory

def eval_spatial_stats(u):
    num_pixels, num_comps = np.prod(u.shape[:-1]), u.shape[-1]
    vert_diffs = np.abs(u[1:,:,:] - u[:-1,:,:])
    horz_diffs = np.abs(u[:,1:,:] - u[:,:-1,:])
    avg_diff = ((np.sum(vert_diffs.reshape((-1, num_comps), order='F'), axis=0) + 
                 np.sum(horz_diffs.reshape((-1, num_comps), order='F'), axis=0)) /
                (np.prod(vert_diffs.shape[:-1]) + np.prod(horz_diffs.shape[:-1])))
    avg_elem = np.mean(np.abs(u.reshape((num_pixels,-1), order='F')), axis=0)
    return avg_diff / avg_elem


def eval_temporal_stats(v):
    return np.mean(np.abs(v[:, :-2] + v[:, 2:] - 2*v[:, 1:-1]), axis=-1) / np.mean(np.abs(v), axis=-1)

def noise_decomp(num_components, params=None):
    """ TESTING: downsampled case """
    E = np.random.randn(params['block_height'], params['block_width'], params['window_length'])
    U = np.zeros((params['block_height'] * params['block_width'] * num_components,), dtype=np.float64)
    V = np.zeros((params['window_length'] * num_components,), dtype=np.float64)

    if params['d_sub'] > 1 or params['t_sub'] > 1:
        # Downsample Simulated Noise
        E_sub = skimage.measure.block_reduce(E, (d_sub, d_sub, t_sub), func=np.mean)

        # Perform Partial Decomposition Without Thresholds W/ Decimated Init
        _ = decimated_decompose(params['block_height'],
                                params['block_width'],
                                params['d_sub'],
                                params['window_length'], 
                                params['t_sub'],
                                np.reshape(E, (-1,)), np.reshape(E_sub, (-1,)),
                                U, V,
                                1e3, 1e3,
                                num_components, num_components,
                                params['max_iters_main'], params['max_iters_init'],
                                params['tol'])
    else:
        # Perform Partial Decomposition Without Thresholds
        _ = decompose(params['block_height'],
                      params['block_width'],
                      params['window_length'], 
                      np.reshape(E, (-1,)),
                      U, V,
                      1e3, 1e3,
                      num_components, num_components,
                      params['max_iters_main'], params['max_iters_init'],
                      params['tol'])

    # Return Summary Stats
    return (eval_spatial_stats(U.reshape((params['block_height'], params['block_width'], -1), order='F')),
            eval_temporal_stats(V.reshape((-1, params['window_length']))))



#  wfield - tools to analyse widefield data - NCAAS related 
# Copyright (C) 2020 Joao Couto - jpcouto@gmail.com
#
#  This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


# This has wrappers and utilities for running or vizualizing data on NeuroCAAS
# Link, references and names should go here.
#
# This has functions to:
#     - Interface with aws to upload data to the cloud
#     - Functions to get experiments from the cloud
#     - Wrappers for data viz and run some parts
#
# Wrapper to run motion correction that takes a filename and kicks out a structure (first step of 'this' blueprint - link)
#



from .utils import *
from .io import load_dat, mmap_dat
from .registration import motion_correct
from .hemocorrection import hemodynamic_correction

def load_and_motion_correct(filename,
                            outputfolder = None,
                            chunksize=1048,
                            nreference = 60,
                            mmap = False,
                            flatten_frames=False,
                            outputdir = None):
    '''
    Motion correction by translation.
    This estimates x and y shifts using phase correlation. And applies shifts to the dataset.
    
    The reference image is the average (nreference frames) from the chunk in the center.

Inputs:
    filename              : the name of the file to perform correction on a data file
    chunksize (int)       : size of the chunks where fft is performed (default 512)
    nreference            : number of frames to take as reference (default 60)
    mmap                  : default False; if True, result is written to disk and returns a memory mapped array
    flatten_frames        : returns [NFRAMES * NCHANNELS, H, W] instead
    outputdir             : folder to write the results to.

Returns:
    motion corrected frames [NFRAMES, NCHANNELS, H, W]
    '''

    if mmap:
        # motion corrected frames overwrite the raw data
        dat = mmap_dat(filename, mode='r+')
    else:
        # this will take a while, loads the entire dataset to memory
        dat = load_dat(filename) 
        
    (yshifts,xshifts),rshifts = motion_correct(dat, chunksize=chunksize,
                                     nreference = nreference,
                                     apply_shifts= True)

    if outputdir is None:
        outputdir = os.path.dirname(filename)
        # save the shifts
        shifts = np.rec.array([yshifts,xshifts],dtype=[('y','float32'),('x','float32')])
        np.save(pjoin(outputdir,'motion_correction_shifts.npy'),shifts)
        np.save(pjoin(outputdir,'motion_correction_rotation.npy'),rshifts)

    try: # don't crash if there are issues plotting
        from .plots import plot_summary_motion_correction 
        import pylab as plt
        plt.matplotlib.style.use('ggplot')
        plot_summary_motion_correction(shifts,localdisk = outputdir);
    except Exception as err:
        print('There was an issue plotting.')
        print(err)

    if mmap:
        del dat
        dat = mmap_dat(dat_path, mode='r')
        
    if flatten_frames:
        return dat.reshape([-1,*dat.shape[-2:]])

    return dat


def dual_color_hemodymamic_correction(U,SVT_470,SVT_405,
                                      output_folder = None,
                                      frame_rate = 30.,
                                      freq_lowpass = 15., 
                                      freq_highpass = 0.1):
    ''' 
    Hemodynamic correction on dual channel data.
    The blue channel is assumed to be the first.
    
    Inputs:
    
    U (array)     : spatial components (H,W,NCOMPONENTS)
    SVT (array)   : temporal components (NCOMPONENTS,NFRAMES*2CHANNELS)
    output_folder : where to store results and figures (default current directory)
    frame_rate    : frame rate of the acquisition 
    freq_lowpass  : lowpass frequency (only for violet - default 15. (none at 30Hz))
    freq_highpass : frequency of the highpass filter, applied to both channels (default is 0.1)

    Returns:
    SVTcorr      : the corrected temporal components (NCOMPONENTS, NFRAMES)
    '''
    
    if output_folder is None:
        output_folder = os.path.abspath(os.path.curdir)
        print('Output not specified, using {0}'.format(output_folder))
    
    SVTcorr, rcoeffs, T = hemodynamic_correction(U,
                                                 SVT_470,
                                                 SVT_405, 
                                                 fs=frame_rate,
                                                 freq_lowpass=freq_lowpass,
                                                 freq_highpass = freq_highpass)        

    np.save(pjoin(output_folder,'rcoeffs.npy'),rcoeffs) # regression coefficients
    np.save(pjoin(output_folder,'T.npy'),T)             # transformation matrix
    np.save(pjoin(output_folder,'SVTcorr.npy'),SVTcorr) # corrected SVT

    try: # don't crash while plotting
        import pylab as plt
        plt.matplotlib.style.use('ggplot')
        from wfield import  plot_summary_hemodynamics_dual_colors
        plot_summary_hemodynamics_dual_colors(rcoeffs,
                                              SVT_470,
                                              SVT_405,
                                              U,
                                              T,
                                              frame_rate=frame_rate,
                                              duration = 12,
                                              outputdir = output_folder);
    except Exception as err:
        print('There was an issue plotting.')
        print(err)
        
    return SVTcorr




#  wfield - tools to analyse widefield data - plotting the results / summary plots 
# Copyright (C) 2020 Joao Couto - jpcouto@gmail.com
#
#  This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .utils import *
from .viz import imshow_noborder

def plot_summary_motion_correction(shifts, localdisk = None,
                                  title = 'Pixel shift (motion correction)'):
    '''
    Shifts are the saved motion correction pixels shifts. 
    shifts          : A rec array with the shifts per channel.
    localdisk       : Output folder (figure will be localdisk/figures/motion_correction.pdf)

    '''
    import pylab as plt
    fig = plt.figure(figsize=[9,5])

    fig.add_subplot(2,1,1)
    # need to make sure this doesnt crash with one channel
    nchannels = shifts['y'].shape[1]
    mi = shifts.view('float32').min() - 1
    ma = shifts.view('float32').max() + 1
    for i in range(nchannels):
        plt.plot(shifts['y'][:,i],label='channel {0}'.format(i+1))
    plt.title(title)
    plt.ylabel('y motion (pixels)')
    plt.legend()
    plt.ylim([mi,ma])

    fig.add_subplot(2,1,2)
    for i in range(nchannels):
        plt.plot(shifts['x'][:,i])
    plt.ylabel('x motion (pixels)')
    plt.xlabel('frame number');        
    plt.ylim([mi,ma])

    if not localdisk is None:
        folder = localdisk
        if not os.path.isdir(folder):
            os.makedirs(folder)
        fig.savefig(pjoin(folder,'motion_correction.pdf'))
    return fig


def plot_summary_hemodynamics_dual_colors(rcoeffs,
                                          SVT_470,
                                          SVT_405,
                                          U,
                                          T,
                                          frame_rate=15.,
                                          duration = 6,
                                          outputdir=None):
    '''
    Approximates what was done in the hemodynamics correction.
    '''

    from skimage.filters import gaussian
    pidx = np.unravel_index(np.argmax(gaussian(rcoeffs,5)),
                           rcoeffs.shape) + np.array([0,-50])

    k,nframes = SVT_470.shape
    u = U[pidx[0]-2:pidx[0]+2,
          pidx[1]-2:pidx[1]+2,:].reshape([-1,k])
    mcoeff = np.mean(rcoeffs[pidx[0]-2:pidx[0]+2,
          pidx[1]-2:pidx[1]+2].reshape([-1]))
    # gets 60 seconds of data
    nframes = int(np.clip(duration*frame_rate,0,SVT_470.shape[1]))
    SVTa = SVT_470#SVT[:,0:nframes:2] 
    SVTb = SVT_405#SVT[:,1:nframes:2]
    # similar proprocessing as for correction
    SVTa = highpass(SVTa,.1,fs = frame_rate)
    SVTb = highpass(SVTb,.1,fs = frame_rate)
    # SVTb = lowpass(SVTb,10., fs = frame_rate)
    # SVTb = lowpass(SVTb,w = 15.,fs = frame_rate/2)
    SVTa = (SVTa.T - np.nanmean(SVTa,axis=1)).T.astype('float32')
    SVTb = (SVTb.T - np.nanmean(SVTb,axis=1)).T.astype('float32')
    SVTb_scaled = np.dot(T,SVTb)
    Ya = np.mean(u@SVTa[:,np.arange(1,nframes)],axis=0)
    Yb = np.mean(u@SVTb[:,np.arange(1,nframes)],axis=0)
    Ybscaled = np.mean(u@SVTb_scaled[:,np.arange(1,nframes)],axis=0)
    Ycorr =  np.mean(u@(SVTa[:,np.arange(1,nframes)]
                        - SVTb_scaled[:,np.arange(1,nframes)]),axis=0)
    
    import pylab as plt
    fig = plt.figure(figsize=[10,4])
    ax = fig.add_axes([.0,.0,.4,1])
    imshow_noborder(rcoeffs, fig = fig, 
               clim = np.percentile(rcoeffs,[20,98]),cmap='inferno')

    plt.colorbar(shrink = .4,label='regression coefficients')

    plt.axhline(y=pidx[0], color="w", linestyle="--",lw=1)
    plt.axvline(x=pidx[1], color="w", linestyle="--",lw=1)

    mi = np.min([np.nanmin(Ya),np.nanmin(Ybscaled)])*1.2
    ma = np.max([np.nanmax(Ya),np.nanmax(Ybscaled)])*1.2

    fig.add_axes([.4,0,.5,.5])
    plt.plot(Ya[:],color='#0066cc',label = 'blue')
    plt.plot(Yb[:],color='#cc00ff',label = 'violet')
    plt.plot(Ybscaled[:],lw = .5,color='k',label = 'scaled violet')
    plt.legend(loc='upper right')

    plt.ylim([mi,ma])
    plt.axis('off')
    fig.add_axes([.4,.5,.5,.5])
    plt.plot(Ycorr[:],'k',label = 'corrected')
    plt.legend(loc='upper right')
    plt.ylim([mi,ma])
    plt.axis('off')
    plt.plot([0,10*frame_rate/2],[mi,mi],'k',clip_on=False)
    plt.plot([0,0],[mi,mi+0.1],'k',clip_on=False)
    plt.text(5*frame_rate/2,mi,'10s',ha='center',va='bottom')
    plt.text(0,mi+0.05,'10% df/f',ha='right',va='center',rotation=90)
    if not outputdir is None:
        folder = outputdir
        if not os.path.isdir(folder):
            os.makedirs(folder)
        fig.savefig(pjoin(folder,'hemodynamic_correction.pdf'))
    return fig
